from Extensions import Extensions
from Agent import get_agents
from Interactions import Interactions
import json
import os
from typing import List, Optional
from transformers import BlipProcessor, BlipForConditionalGeneration
import torch
from PIL import Image
import requests
import subprocess


class agixt_agent(Extensions):
    def __init__(self, **kwargs):
        agents = get_agents()
        self.commands = {
            "Evaluate Code": self.evaluate_code,
            "Analyze Pull Request": self.analyze_pull_request,
            "Perform Automated Testing": self.perform_automated_testing,
            "Run CI-CD Pipeline": self.run_ci_cd_pipeline,
            "Improve Code": self.improve_code,
            "Write Tests": self.write_tests,
            "Create a new command": self.create_command,
            "Describe Image": self.describe_image,
            "Execute Python Code": self.execute_python_code,
            "Get Python Code from Response": self.get_python_code_from_response,
        }
        if agents != None:
            for agent in agents:
                if "name" in agent:
                    name = f" AI Agent {agent['name']}"
                    self.commands.update(
                        {
                            f"Ask{name}": self.ask,
                            f"Instruct{name}": self.instruct,
                        }
                    )

    def command_exists(self, file_name: str) -> bool:
        return os.path.exists(f"commands/{file_name}.py")

    async def create_command(
        self, function_description: str, agent: str = "AGiXT"
    ) -> List[str]:
        with open(f"prompts/Create New Command.txt", "r") as f:
            prompt = f.read()
        prompt = prompt.replace("{{NEW_FUNCTION_DESCRIPTION}}", function_description)
        response = await Interactions(agent_name=agent).run(user_input=prompt)
        file_name = response.split("class ")[1].split("(")[0]
        code = code.replace("```", "")

        if not self.command_exists(file_name):
            with open(f"commands/{file_name}.py", "w") as f:
                f.write(code)
            return f"Created new command: {file_name}."
        else:
            return f"Command {file_name} already exists. No changes were made."

    async def evaluate_code(self, code: str, agent: str = "AGiXT") -> List[str]:
        args = [code]
        function_string = "def analyze_code(code: str) -> List[str]:"
        description_string = "Analyzes the given code and returns a list of suggestions for improvements."
        prompt = f"You are now the following python function: ```# {description_string}\n{function_string}```\n\nOnly respond with your `return` value. Args: {args}"
        return await Interactions(agent_name=agent).run(user_input=prompt)

    async def analyze_pull_request(
        self, pr_url: str, agent: str = "AGiXT"
    ) -> List[str]:
        args = [pr_url]
        function_string = "def analyze_pr(pr_url: str) -> List[str]:"
        description_string = "Analyzes the given pull request and returns a list of suggestions for improvements."
        prompt = f"You are now the following python function: ```# {description_string}\n{function_string}```\n\nOnly respond with your `return` value. Args: {args}"
        return await Interactions(agent_name=agent).run(user_input=prompt)

    async def perform_automated_testing(
        self, test_url: str, agent: str = "AGiXT"
    ) -> List[str]:
        args = [test_url]
        function_string = "def perform_testing(test_url: str) -> List[str]:"
        description_string = "Performs automated testing using AI-driven tools and returns a list of test results."
        prompt = f"You are now the following python function: ```# {description_string}\n{function_string}```\n\nOnly respond with your `return` value. Args: {args}"
        return await Interactions(agent_name=agent).run(user_input=prompt)

    async def improve_code(
        self, suggestions: List[str], code: str, agent: str = "AGiXT"
    ) -> str:
        args = [json.dumps(suggestions), code]
        function_string = (
            "def generate_improved_code(suggestions: List[str], code: str) -> str:"
        )
        description_string = "Improves the provided code based on the suggestions provided, making no other changes."
        prompt = f"You are now the following python function: ```# {description_string}\n{function_string}```\n\nOnly respond with your `return` value. Args: {args}"
        return await Interactions(agent_name=agent).run(user_input=prompt)

    async def write_tests(
        self,
        code: str,
        focus: Optional[List[str]] = None,
        agent: str = "AGiXT",
    ) -> str:
        args = [code, json.dumps(focus) if focus else "None"]
        function_string = "def create_test_cases(code: str, focus: Optional[List[str]] = None) -> str:"
        description_string = "Generates test cases for the existing code, focusing on specific areas if required."
        prompt = f"You are now the following python function: ```# {description_string}\n{function_string}```\n\nOnly respond with your `return` value. Args: {args}"
        return await Interactions(agent_name=agent).run(user_input=prompt)

    async def run_ci_cd_pipeline(self, repo_url: str, agent: str = "AGiXT") -> str:
        args = [repo_url]
        function_string = "def run_pipeline(repo_url: str) -> str:"
        description_string = (
            "Runs the entire CI/CD pipeline for the given repository URL."
        )
        prompt = f"You are now the following python function: ```# {description_string}\n{function_string}```\n\nOnly respond with your `return` value. Args: {args}"
        return await Interactions(agent_name=agent).run(user_input=prompt)

    async def ask(self, user_input: str, agent: str = "AGiXT") -> str:
        response = await Interactions(agent_name=agent).run(
            user_input=user_input, prompt="chat", websearch=True, websearch_depth=4
        )
        return response

    async def instruct(self, user_input: str, agent: str = "AGiXT") -> str:
        response = await Interactions(agent_name=agent).run(
            user_input=user_input, prompt="instruct", websearch=True, websearch_depth=8
        )
        return response

    async def describe_image(self, image_url):
        """
        Describe an image using FuseCap.
        """
        if image_url:
            processor = BlipProcessor.from_pretrained("noamrot/FuseCap")
            model = BlipForConditionalGeneration.from_pretrained("noamrot/FuseCap")

            # Define the device to run the model on (CPU or GPU)

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            model.to(device)
            raw_image = Image.open(requests.get(image_url, stream=True).raw).convert(
                "RGB"
            )

            # Generate a caption for the image using FuseCap
            text = "a picture of "
            inputs = processor(raw_image, text, return_tensors="pt").to(device)
            out = model.generate(max_length=20, temperature=0.7, **inputs, num_beams=3)
            caption = processor.decode(out[0], skip_special_tokens=True)

            # Return the caption
            return caption

    async def get_python_code_from_response(self, response: str):
        if "```python" in response:
            response = response.split("```python")[1].split("```")[0]
        return response

    async def execute_python_code(self, code: str) -> str:
        code = await self.get_python_code_from_response(code)
        # Create the WORKSPACE directory if it doesn't exist
        workspace_dir = os.path.join(os.getcwd(), "WORKSPACE")
        os.makedirs(workspace_dir, exist_ok=True)

        # Create a temporary Python file in the WORKSPACE directory
        temp_file = os.path.join(workspace_dir, "temp.py")
        with open(temp_file, "w") as f:
            f.write(code)

        try:
            # Execute the Python script and capture its output
            response = subprocess.check_output(
                ["python", temp_file], stderr=subprocess.STDOUT
            )
        except subprocess.CalledProcessError as e:
            # If the script raises an exception, return the error message
            response = e.output

        # Delete the temporary Python file
        os.remove(temp_file)

        # Decode the output from bytes to string and return it
        return response.decode("utf-8")



import poe

"""
Models for Poe:

self.client.bot_names = {
  "capybara": "Sage",
  "a2": "Claude-instant",
  "nutria": "Dragonfly",
  "a2_100k": "Claude-instant-100k",
  "beaver": "GPT-4",
  "chinchilla": "ChatGPT",
  "a2_2": "Claude+"
}
Notes for Free accounts:

Claude+ (a2_2) has a limit of 3 messages per day. 
GPT-4 (beaver) has a limit of 1 message per day. 
Claude-instant-100k (c2_100k) is completely inaccessible for free accounts. 
For all the other chatbots, there seems to be a rate limit of 10 messages per minute.
"""


class PoeProvider:
    def __init__(self, POE_TOKEN: str = "", AI_MODEL: str = "chinchilla", **kwargs):
        self.requirements = ["poe-api"]
        self.POE_TOKEN = POE_TOKEN
        self.AI_MODEL = AI_MODEL.lower()

    async def instruct(self, prompt, tokens: int = 0):
        try:
            client = poe.Client(token=self.POE_TOKEN)
            if self.AI_MODEL not in client.bot_names:
                try:
                    self.AI_MODEL = client.get_bot_by_codename(self.AI_MODEL)
                except:
                    raise Exception(f"Invalid AI Model: {self.AI_MODEL}")
            for chunk in client.send_message(chatbot=self.AI_MODEL, message=prompt):
                pass
            response = chunk["text"].replace("\n", "\n")
            return response
        except Exception as e:
            return f"Poe Error: {e}"



import json
import streamlit as st
from ApiClient import ApiClient
from components.verify_backend import verify_backend
from components.docs import agixt_docs
from components.chain import modify_chain

verify_backend()

st.set_page_config(
    page_title="Chain Management",
    page_icon=":chains:",
    layout="wide",
)

agixt_docs()
st.session_state = {}
chain_names = ApiClient.get_chains()
agents = ApiClient.get_agents()
st.header("Chain Management")
chain_action = st.selectbox("Action", ["Create Chain", "Modify Chain", "Delete Chain"])

if chain_action == "Create Chain":
    chain_name = st.text_input("Chain Name")
else:
    chain_name = st.selectbox("Chains", options=chain_names)

if chain_action == "Create Chain":
    action_button = st.button("Create New Chain")
    # Import Chain
    chain_file = st.file_uploader("Import Chain", type=["json"])
    if chain_file:
        chain_name = chain_file.name.split(".")[0]
        chain_content = chain_file.read().decode("utf-8")
        steps = json.loads(chain_content)
        ApiClient.import_chain(chain_name=chain_name, steps=steps)
        st.success(f"Chain '{chain_name}' added.")
        chain_file = None
    if action_button:
        if chain_name:
            ApiClient.add_chain(chain_name=chain_name)
            st.success(f"Chain '{chain_name}' created.")
            st.experimental_rerun()
        else:
            st.error("Chain name is required.")

elif chain_action == "Delete Chain":
    action_button = st.button("Delete Chain")
    if action_button:
        if chain_name:
            ApiClient.delete_chain(chain_name=chain_name)
            st.success(f"Chain '{chain_name}' deleted.")
            st.experimental_rerun()
        else:
            st.error("Chain name is required.")

elif chain_action == "Modify Chain":
    if chain_name:
        chain = modify_chain(chain_name=chain_name, agents=agents)
    else:
        st.warning("Please select a chain to manage steps.")

st.markdown("### Predefined Injection Variables")
st.markdown(
    """
    Any of these variables can be used in command arguments or prompt arguments to inject data into the prompt. These can also be used inside of any Custom Prompt.
- `{agent_name}` will cause the agent name to be injected.
- `{context}` will cause the current context from memory to be injected. This will only work if you have `{user_input}` in your prompt arguments for the memory search. (Only applies to prompts but is still a reserved variable name.)
- `{date}` will cause the current date and timestamp to be injected.
- `{COMMANDS}` will cause the available commands list to be injected and for automatic commands execution from the agent based on its suggestions.
- `{command_list}` will cause the available commands list to be injected, but will not execute any commands the AI chooses. Useful on validation steps.
- `{STEPx}` will cause the step `x` response from a chain to be injected. For example, `{STEP1}` will inject the first step's response in a chain.
"""
)



import os
import json
import glob
import shutil
import importlib
import yaml
import time
from pathlib import Path
from inspect import signature, Parameter
from provider import Provider
from Memories import Memories
from Extensions import Extensions
from datetime import datetime

DEFAULT_SETTINGS = {
    "provider": "gpt4free",
    "AI_MODEL": "gpt-3.5-turbo",
    "AI_TEMPERATURE": "0.7",
    "MAX_TOKENS": "4096",
    "embedder": "default",
    "autonomous_execution": "false",
}


def get_agent_file_paths(agent_name):
    base_path = os.path.join(os.getcwd(), "agents")
    folder_path = os.path.normpath(os.path.join(base_path, agent_name))
    config_path = os.path.normpath(os.path.join(folder_path, "config.json"))
    history_path = os.path.normpath(os.path.join(folder_path, "history.yaml"))
    if not config_path.startswith(base_path) or not folder_path.startswith(base_path):
        raise ValueError("Invalid path, agent name must not contain slashes.")
    if not os.path.exists(folder_path):
        os.mkdir(folder_path)
    return config_path, history_path, folder_path


def add_agent(agent_name, provider_settings=None, commands={}):
    if not agent_name:
        return "Agent name cannot be empty."
    provider_settings = (
        DEFAULT_SETTINGS
        if not provider_settings or provider_settings == {}
        else provider_settings
    )
    config_path, history_path, folder_path = get_agent_file_paths(agent_name=agent_name)
    if provider_settings is None or provider_settings == "" or provider_settings == {}:
        provider_settings = DEFAULT_SETTINGS
    settings = json.dumps(
        {
            "commands": commands,
            "settings": provider_settings,
        }
    )
    # Write the settings to the agent config file
    with open(config_path, "w") as f:
        f.write(settings)
    with open(history_path, "w") as f:
        f.write("")
    return {"message": f"Agent {agent_name} created."}


def delete_agent(agent_name):
    config_path, history_path, folder_path = get_agent_file_paths(agent_name=agent_name)
    try:
        if os.path.exists(folder_path):
            shutil.rmtree(folder_path)
        return {"message": f"Agent {agent_name} deleted."}, 200
    except:
        return {"message": f"Agent {agent_name} could not be deleted."}, 400


def rename_agent(agent_name, new_name):
    config_path, history_path, folder_path = get_agent_file_paths(agent_name=agent_name)
    base_path = os.path.join(os.getcwd(), "agents")
    new_agent_folder = os.path.normpath(os.path.join(base_path, new_name))
    if not new_agent_folder.startswith(base_path):
        raise ValueError("Invalid path, agent name must not contain slashes.")

    if os.path.exists(folder_path):
        # Check if the new name is already taken
        if os.path.exists(new_agent_folder):
            # Add a number to the end of the new name
            i = 1
            while os.path.exists(new_agent_folder):
                i += 1
                new_name = f"{new_name}_{i}"
                new_agent_folder = os.path.normpath(os.path.join(base_path, new_name))
            if not new_agent_folder.startswith(base_path):
                raise ValueError("Invalid path, agent name must not contain slashes.")
        os.rename(folder_path, new_agent_folder)
        return {"message": f"Agent {agent_name} renamed to {new_name}."}, 200


def get_agents():
    agents_dir = "agents"
    if not os.path.exists(agents_dir):
        os.makedirs(agents_dir)
    agents = [
        dir_name
        for dir_name in os.listdir(agents_dir)
        if os.path.isdir(os.path.join(agents_dir, dir_name))
    ]
    output = []
    if agents:
        for agent in agents:
            output.append({"name": agent, "status": False})
    return output


class Agent:
    def __init__(self, agent_name=None):
        self.agent_name = agent_name if agent_name is not None else "AGiXT"
        self.config_path, self.history_file, self.folder_path = get_agent_file_paths(
            agent_name=self.agent_name
        )
        self.AGENT_CONFIG = self.get_agent_config()
        if "settings" in self.AGENT_CONFIG:
            self.PROVIDER_SETTINGS = self.AGENT_CONFIG["settings"]
            if "provider" in self.PROVIDER_SETTINGS:
                self.AI_PROVIDER = self.PROVIDER_SETTINGS["provider"]
                self.PROVIDER = Provider(self.AI_PROVIDER, **self.PROVIDER_SETTINGS)
                self._load_agent_config_keys(
                    ["AI_MODEL", "AI_TEMPERATURE", "MAX_TOKENS", "autonomous_execution"]
                )
            if "AI_MODEL" in self.PROVIDER_SETTINGS:
                self.AI_MODEL = self.PROVIDER_SETTINGS["AI_MODEL"]
                if self.AI_MODEL == "":
                    self.AI_MODEL = "default"
            else:
                self.AI_MODEL = "openassistant"
            if "embedder" in self.PROVIDER_SETTINGS:
                self.EMBEDDER = self.PROVIDER_SETTINGS["embedder"]
            else:
                if self.AI_PROVIDER == "openai":
                    self.EMBEDDER = "openai"
                else:
                    self.EMBEDDER = "default"
            if "MAX_TOKENS" in self.PROVIDER_SETTINGS:
                self.MAX_TOKENS = self.PROVIDER_SETTINGS["MAX_TOKENS"]
            else:
                self.MAX_TOKENS = 4000
            if "LOG_REQUESTS" in self.PROVIDER_SETTINGS:
                self.LOG_REQUESTS = self.PROVIDER_SETTINGS["LOG_REQUESTS"]
            else:
                self.LOG_REQUESTS = True
            if "autonomous_execution" in self.PROVIDER_SETTINGS:
                self.AUTONOMOUS_EXECUTION = self.PROVIDER_SETTINGS[
                    "autonomous_execution"
                ]
            else:
                self.AUTONOMOUS_EXECUTION = True
            self.commands = self.load_commands()
            self.available_commands = Extensions(
                agent_config=self.AGENT_CONFIG
            ).get_available_commands()
            self.clean_agent_config_commands()
            self.history = self.load_history()
            self.agent_instances = {}
            self.agent_config = self.load_agent_config()
            if self.LOG_REQUESTS:
                Path(
                    os.path.normpath(
                        os.path.join(
                            self.folder_path,
                            "requests",
                        )
                    )
                ).mkdir(parents=True, exist_ok=True)

    def get_memories(self):
        return Memories(self.agent_name, self.AGENT_CONFIG)

    async def execute(self, command_name, command_args):
        return await Extensions(agent_config=self.AGENT_CONFIG).execute_command(
            command_name=command_name, command_args=command_args
        )

    async def instruct(self, prompt, tokens):
        if not prompt:
            return ""
        answer = await self.PROVIDER.instruct(prompt=prompt, tokens=tokens)
        if self.LOG_REQUESTS:
            log_file = os.path.join(
                "agents", self.agent_name, "requests", f"{time.time()}.txt"
            )
            with open(
                log_file,
                "a" if os.path.exists(log_file) else "w",
                encoding="utf-8",
            ) as f:
                f.write(f"{prompt}\n{answer}")
        return answer

    def _load_agent_config_keys(self, keys):
        for key in keys:
            if key in self.AGENT_CONFIG:
                setattr(self, key, self.AGENT_CONFIG[key])

    def clean_agent_config_commands(self):
        for command in self.commands:
            friendly_name = command[0]
            if friendly_name not in self.AGENT_CONFIG["commands"]:
                self.AGENT_CONFIG["commands"][friendly_name] = False
        for command in list(self.AGENT_CONFIG["commands"]):
            if command not in [cmd[0] for cmd in self.commands]:
                del self.AGENT_CONFIG["commands"][command]
        with open(self.config_path, "w") as f:
            json.dump(self.AGENT_CONFIG, f)

    def get_commands_string(self):
        if len(self.available_commands) == 0:
            return None

        enabled_commands = filter(
            lambda command: command.get("enabled", True), self.available_commands
        )
        if not enabled_commands:
            return None

        friendly_names = map(
            lambda command: f"`{command['friendly_name']}` - Arguments: {command['args']}",
            enabled_commands,
        )
        command_list = "\n".join(friendly_names)
        return f"Commands Available To Complete Task:\n{command_list}\n\n"

    def get_provider(self):
        config_file = self.get_agent_config()
        if "provider" in config_file:
            return config_file["provider"]
        else:
            return "openai"

    def get_command_params(self, func):
        params = {}
        sig = signature(func)
        for name, param in sig.parameters.items():
            if param.default == Parameter.empty:
                params[name] = None
            else:
                params[name] = param.default
        return params

    def load_commands(self):
        commands = []
        command_files = glob.glob("extensions/*.py")
        for command_file in command_files:
            module_name = os.path.splitext(os.path.basename(command_file))[0]
            module = importlib.import_module(f"extensions.{module_name}")
            command_class = getattr(module, module_name.lower())()
            if hasattr(command_class, "commands"):
                for command_name, command_function in command_class.commands.items():
                    params = self.get_command_params(command_function)
                    commands.append((command_name, command_function.__name__, params))
        return commands

    def create_agent_config_file(self, provider_settings, commands):
        if (
            provider_settings is None
            or provider_settings == ""
            or provider_settings == {}
        ):
            provider_settings = DEFAULT_SETTINGS
        settings = json.dumps(
            {
                "commands": commands,
                "settings": provider_settings,
            }
        )

        # Check and create agent directory if it doesn't exist
        if not os.path.exists(os.path.join("agents", self.agent_name)):
            os.makedirs(os.path.join("agents", self.agent_name))

        # Write the settings to the agent config file
        with open(self.config_path, "w") as f:
            f.write(settings)

        return self.config_path

    def load_agent_config(self):
        try:
            with open(self.config_path) as agent_config:
                try:
                    agent_config_data = json.load(agent_config)
                    return agent_config_data
                except json.JSONDecodeError:
                    agent_config_data = {}
                    # Populate the agent_config with all commands enabled
                    agent_config_data["commands"] = {
                        command_name: "false"
                        for command_name, _, _ in self.load_commands(self.agent_name)
                    }
                    agent_config_data["settings"] = DEFAULT_SETTINGS
                    # Save the updated agent_config to the file
                    with open(self.config_path, "w") as agent_config_file:
                        json.dump(agent_config_data, agent_config_file)
                    return agent_config_data
        except:
            with open(self.config_path, "w") as f:
                f.write(
                    json.dumps(
                        {
                            "commands": {
                                command_name: "false"
                                for command_name, _, _ in self.load_commands()
                            },
                            "settings": DEFAULT_SETTINGS,
                        }
                    )
                )
        return agent_config_data

    def get_agent_config(self):
        while True:
            if os.path.exists(self.config_path):
                try:
                    with open(self.config_path, "r") as f:
                        file_content = f.read().strip()
                        if file_content:
                            return json.loads(file_content)
                except:
                    None
            add_agent(agent_name=self.agent_name)
            return self.get_agent_config()

    def update_agent_config(self, new_config, config_key):
        if os.path.exists(self.config_path):
            with open(self.config_path, "r") as f:
                current_config = json.load(f)

            # Ensure the config_key is present in the current configuration
            if config_key not in current_config:
                current_config[config_key] = {}

            # Update the specified key with new_config while preserving other keys and values
            for key, value in new_config.items():
                current_config[config_key][key] = value

            # Save the updated configuration back to the file
            with open(self.config_path, "w") as f:
                json.dump(current_config, f)
            return f"Agent {self.agent_name} configuration updated."
        else:
            return f"Agent {self.agent_name} configuration not found."

    def get_history(self):
        if not os.path.exists(self.history_file):
            with open(self.history_file, "w") as f:
                f.write("")
            return []
        try:
            with open(self.history_file, "r") as f:
                yaml_history = yaml.safe_load(f)
            if "interactions" in yaml_history:
                return yaml_history["interactions"]
            return []
        except:
            return []

    def wipe_agent_memories(self):
        memories_folder = os.path.normpath(
            os.path.join(os.getcwd(), self.agent_name, "memories")
        )
        if not memories_folder.startswith(os.getcwd()):
            raise ValueError("Invalid path, agent name must not contain slashes.")

        if os.path.exists(memories_folder):
            shutil.rmtree(memories_folder)

    def load_history(self):
        if os.path.exists(self.history_file):
            with open(self.history_file, "r") as file:
                memory = yaml.safe_load(file)
        else:
            with open(self.history_file, "w") as file:
                yaml.safe_dump({"interactions": []}, file)
            memory = {"interactions": []}
        return memory

    def save_history(self):
        with open(self.history_file, "w") as file:
            yaml.safe_dump(self.history, file)

    def log_interaction(self, role: str, message: str):
        if self.history is None:
            self.history = {"interactions": []}
        self.history["interactions"].append(
            {
                "role": role,
                "message": message,
                "timestamp": datetime.now().strftime("%B %d, %Y %I:%M %p"),
            }
        )
        self.save_history()

    def delete_history(self):
        try:
            self.history = {"interactions": []}
            self.save_history()
            return "History deleted."
        except:
            return "History not found."

    def delete_history_message(self, message: str):
        try:
            self.history["interactions"] = [
                interaction
                for interaction in self.history["interactions"]
                if interaction["message"] != message
            ]
            self.save_history()
            return "Message deleted."
        except:
            return "Message not found."



import streamlit as st
from components.selectors import agent_selection
from ApiClient import ApiClient
from components.learning import learning_page
from components.history import get_history
from components.verify_backend import verify_backend
from components.docs import agixt_docs

st.set_page_config(
    page_title="Interact with Agents",
    page_icon=":speech_balloon:",
    layout="wide",
)


verify_backend()


agixt_docs()

st.header("Interact with Agents")
# Create an instance of the API Client
api_client = ApiClient()

# Fetch available prompts
prompts = api_client.get_prompts()

# Add a dropdown to select a mode
mode = st.selectbox("Select Mode", ["Prompt", "Chat", "Instruct", "Learning", "Chains"])

if "chat_history" not in st.session_state:
    st.session_state["chat_history"] = ""

agent_name = agent_selection() if mode != "Chains" else None


with st.container():
    if agent_name:
        st.session_state["chat_history"] = get_history(agent_name=agent_name)


# If the user selects Prompt, then show the prompt functionality
if mode == "Prompt":
    st.markdown("### Choose an Agent and Prompt")
    # Add a dropdown to select a prompt
    prompt_name = st.selectbox("Choose a prompt", prompts)
    # Fetch arguments for the selected prompt
    prompt_args = api_client.get_prompt_args(prompt_name=prompt_name)

    # Add input fields for prompt arguments
    st.markdown("### Prompt Variables")
    prompt_args_values = {}
    skip_args = ["command_list", "context", "COMMANDS", "date"]
    for arg in prompt_args:
        if arg not in skip_args:
            prompt_args_values[arg] = st.text_area(arg)

    # Add a checkbox for websearch option
    websearch = st.checkbox("Enable websearch")
    websearch_depth = (
        3 if websearch else 0
    )  # Default depth is 3 if websearch is enabled

    # Add an input field for websearch depth if websearch is enabled
    if websearch:
        websearch_depth = st.number_input("Websearch depth", min_value=1, value=3)

    # Add an input field for context_results if 'task' is in prompt_args
    context_results = 0
    if "task" in prompt_args and "context" in prompt_args:
        context_results = st.number_input("Context results", min_value=1, value=5)

    # Button to execute the prompt
    if st.button("Execute"):
        # Call the prompt_agent function
        agent_prompt_resp = api_client.prompt_agent(
            agent_name=agent_name,
            prompt_name=prompt_name,
            prompt_args=prompt_args_values,
            websearch=websearch,
            websearch_depth=websearch_depth,
            context_results=context_results,
            shots=1,
        )

        # Print the response
        st.write(f"{agent_name}: {agent_prompt_resp}")

if mode == "Chat":
    st.markdown("### Choose an Agent to Chat With")
    smart_chat_toggle = st.checkbox("Enable Smart Chat")
    chat_prompt = st.text_area("Enter your message", key="chat_prompt")
    send_button = st.button("Send Message")

    if send_button:
        if agent_name and chat_prompt:
            with st.spinner("Thinking, please wait..."):
                if smart_chat_toggle:
                    response = ApiClient.smartchat(
                        agent_name=agent_name,
                        prompt=chat_prompt,
                        shots=3,
                    )
                else:
                    response = ApiClient.chat(agent_name=agent_name, prompt=chat_prompt)
                if response:
                    st.experimental_rerun()


if mode == "Instruct":
    st.markdown("### Choose an Agent to Instruct")
    smart_instruct_toggle = st.checkbox("Enable Smart Instruct")
    instruct_prompt = st.text_area("Enter your instruction", key="instruct_prompt")
    send_button = st.button("Send Message")

    if send_button:
        if agent_name and instruct_prompt:
            with st.spinner("Thinking, please wait..."):
                if smart_instruct_toggle:
                    response = ApiClient.smartinstruct(
                        agent_name=agent_name,
                        prompt=instruct_prompt,
                        shots=3,
                    )
                else:
                    response = ApiClient.instruct(
                        agent_name=agent_name, prompt=instruct_prompt
                    )
            if response:
                st.experimental_rerun()

if mode == "Learning":
    if agent_name:
        learning_page(agent_name)

if mode == "Chains":
    st.markdown("### Chain to Run")
    chain_names = ApiClient.get_chains()
    chain_action = "Run Chain"
    chain_name = st.selectbox("Chains", chain_names)
    from_step = st.number_input("Start from Step", min_value=1, value=1)
    all_responses = st.checkbox(
        "Show All Responses (If not checked, you will only be shown the last step's response in the chain when done.)"
    )
    user_input = st.text_area("User Input")
    # Need a checkbox for agent override
    agent_override = st.checkbox("Override Agent")
    if agent_override:
        agent_name = agent_selection()
    else:
        agent_name = ""
    if st.button("Run Chain"):
        if chain_name:
            if chain_action == "Run Chain":
                responses = ApiClient.run_chain(
                    chain_name=chain_name,
                    user_input=user_input,
                    agent_name=agent_name,
                    all_responses=all_responses,
                    from_step=from_step,
                )
                st.success(f"Chain '{chain_name}' executed.")
                st.write(responses)
        else:
            st.error("Chain name is required.")



import streamlit as st
from ApiClient import ApiClient
from components.verify_backend import verify_backend
from components.docs import agixt_docs

verify_backend()

st.set_page_config(
    page_title="Prompt Templates",
    page_icon=":scroll:",
    layout="wide",
)
agixt_docs()

st.header("Prompt Templates")

prompt_list = ApiClient.get_prompts()

action = st.selectbox("Action", ["Create New Prompt", "Modify Prompt", "Delete Prompt"])

if action == "Create New Prompt":
    # Import prompt button
    prompt_file = st.file_uploader("Import Prompt", type=["txt"])
    if prompt_file:
        prompt_name = prompt_file.name.split(".")[0]
        prompt_content = prompt_file.read().decode("utf-8")
        ApiClient.add_prompt(prompt_name=prompt_name, prompt=prompt_content)
        st.success(f"Prompt '{prompt_name}' added.")
    prompt_name = st.text_input("Prompt Name")
    prompt_content = st.text_area("Prompt Content", height=300)

elif action == "Modify Prompt":
    prompt_name = st.selectbox("Existing Prompts", prompt_list)
    prompt_content = st.text_area(
        "Prompt Content",
        ApiClient.get_prompt(prompt_name=prompt_name) if prompt_name else "",
        height=300,
    )
    export_button = st.download_button(
        "Export Prompt", data=prompt_content, file_name=f"{prompt_name}.txt"
    )
elif action == "Delete Prompt":
    prompt_name = st.selectbox("Existing Prompts", prompt_list)
    prompt_content = None

if st.button("Perform Action"):
    if prompt_name and (prompt_content or action == "Delete Prompt"):
        if action == "Create New Prompt":
            ApiClient.add_prompt(prompt_name=prompt_name, prompt=prompt_content)
            st.success(f"Prompt '{prompt_name}' added.")
        elif action == "Modify Prompt":
            ApiClient.update_prompt(prompt_name=prompt_name, prompt=prompt_content)
            st.success(f"Prompt '{prompt_name}' updated.")
        elif action == "Delete Prompt":
            ApiClient.delete_prompt(prompt_name)
            st.success(f"Prompt '{prompt_name}' deleted.")
    else:
        st.error("Prompt name and content are required.")

st.markdown("### Usage Instructions")
st.markdown(
    """
To create dynamic prompts that can have user inputs, you can use curly braces `{}` in your prompt content. 
Anything between the curly braces will be considered as an input field. For example:

```python
"Hello, my name is {name} and I'm {age} years old."
```
In the above prompt, `name` and `age` will be the input arguments. These arguments can be used in chains.
"""
)
st.markdown("### Predefined Injection Variables")
st.markdown(
    """
- `{agent_name}` will cause the agent name to be injected.
- `{context}` will cause the current context from memory to be injected. This will only work if you have `{user_input}` in your prompt arguments for the memory search.
- `{date}` will cause the current date and timestamp to be injected.
- `{COMMANDS}` will cause the available commands list to be injected and for automatic commands execution from the agent based on its suggestions.
- `{command_list}` will cause the available commands list to be injected, but will not execute any commands the AI chooses. Useful on validation steps.
- `{STEPx}` will cause the step `x` response from a chain to be injected. For example, `{STEP1}` will inject the first step's response in a chain.
"""
)



import os
import json
import streamlit as st
from ApiClient import ApiClient
from components.selectors import agent_selection
from components.verify_backend import verify_backend
from components.docs import agixt_docs

verify_backend()

st.set_page_config(
    page_title="Agent Settings",
    page_icon=":hammer_and_wrench:",
    layout="wide",
)
agixt_docs()


@st.cache_data
def get_providers():
    return ApiClient.get_providers()


@st.cache_data
def get_embed_providers():
    return ApiClient.get_embed_providers()


@st.cache_data
def provider_settings(provider_name: str):
    return ApiClient.get_provider_settings(provider_name)


@st.cache_data
def get_extension_settings():
    return ApiClient.get_extension_settings()


providers = get_providers()
embedders = get_embed_providers()
extension_setting_keys = get_extension_settings()


def render_provider_settings(agent_settings, provider_name: str):
    try:
        required_settings = provider_settings(provider_name)
    except (TypeError, ValueError):
        st.error(
            f"Error loading provider settings: expected a list or a dictionary, but got {required_settings}"
        )
        return {}
    rendered_settings = {}

    if not isinstance(required_settings, (list, dict)):
        st.error(
            f"Error loading provider settings: expected a list or a dictionary, but got {required_settings}"
        )
        return rendered_settings

    if isinstance(required_settings, dict):
        required_settings = list(required_settings.keys())

    for key in required_settings:
        if key in agent_settings:
            default_value = agent_settings[key]
        else:
            default_value = ""

        user_val = st.text_input(key, value=default_value)
        rendered_settings[key] = user_val

    if "LOG_REQUESTS" in agent_settings:
        value = agent_settings["LOG_REQUESTS"]
    else:
        value = False
    rendered_settings["LOG_REQUESTS"] = st.checkbox(
        "Log requests to files", key="LOG_REQUESTS", value=value
    )

    return rendered_settings


st.header("Agent Settings")
agent_name = agent_selection()

if "new_agent_name" not in st.session_state:
    st.session_state["new_agent_name"] = ""

# Add an input field for the new agent's name
new_agent = False

# Check if a new agent has been added and reset the session state variable
if (
    st.session_state["new_agent_name"]
    and st.session_state["new_agent_name"] != agent_name
):
    st.session_state["new_agent_name"] = ""

if not agent_name:
    agent_file = st.file_uploader("Import Agent", type=["json"])
    if agent_file:
        agent_name = agent_file.name.split(".")[0]
        agent_settings = agent_file.read().decode("utf-8")
        agent_config = json.loads(agent_settings)
        ApiClient.import_agent(
            agent_name=agent_name,
            settings=agent_config["settings"],
            commands=agent_config["commands"],
        )
        st.success(f"Agent '{agent_name}' imported.")
    new_agent_name = st.text_input("New Agent Name")

    # Add an "Add Agent" button
    add_agent_button = st.button("Add Agent")

    # If the "Add Agent" button is clicked, create a new agent config file
    if add_agent_button:
        if new_agent_name:
            try:
                ApiClient.add_agent(new_agent_name, {})
                st.success(f"Agent '{new_agent_name}' added.")
                agent_name = new_agent_name
                with open(os.path.join("session.txt"), "w") as f:
                    f.write(agent_name)
                st.session_state["new_agent_name"] = agent_name
                st.experimental_rerun()  # Rerun the app to update the agent list
            except Exception as e:
                st.error(f"Error adding agent: {str(e)}")
        else:
            st.error("New agent name is required.")
    new_agent = True

if agent_name and not new_agent:
    try:
        agent_config = ApiClient.get_agentconfig(agent_name=agent_name)
        export_button = st.download_button(
            "Export Agent Config",
            data=json.dumps(agent_config, indent=4),
            file_name=f"{agent_name}.json",
            mime="application/json",
        )
        agent_settings = agent_config.get("settings", {})
        provider_name = agent_settings.get("provider", "")
        provider_name = st.selectbox(
            "Select Provider",
            providers,
            index=providers.index(provider_name) if provider_name in providers else 0,
        )

        agent_settings[
            "provider"
        ] = provider_name  # Update the agent_settings with the selected provider

        embedder_name = agent_settings.get("embedder", "")
        embedder_name = st.selectbox(
            "Select Embedder",
            embedders,
            index=embedders.index(embedder_name) if embedder_name in embedders else 0,
        )

        agent_settings[
            "embedder"
        ] = embedder_name  # Update the agent_settings with the selected embedder

        if provider_name:
            provider_settings = render_provider_settings(agent_settings, provider_name)
            agent_settings.update(provider_settings)

        def render_extension_settings(extension_settings, agent_settings):
            rendered_settings = {}

            for extension, settings in extension_settings.items():
                st.subheader(f"{extension}")
                for key, val in settings.items():
                    if key in agent_settings:
                        default_value = agent_settings[key]
                    else:
                        default_value = val if val else ""

                    user_val = st.text_input(
                        key, value=default_value, key=f"{extension}_{key}"
                    )

                    # Check if the user value exists before saving the setting
                    if user_val:
                        rendered_settings[key] = user_val

            return rendered_settings

        with st.form(key="update_agent_settings_form"):
            update_agent_settings_button = st.form_submit_button(
                "Update Agent Settings"
            )
            wipe_memories_button = st.form_submit_button("Wipe Agent Memories")
            delete_agent_button = st.form_submit_button("Delete Agent")
        st.subheader("Extension Settings")
        with st.form("extension_settings"):
            extension_settings = render_extension_settings(
                extension_setting_keys, agent_settings
            )

            # Update the extension settings in the agent_settings directly
            agent_settings.update(extension_settings)

            st.subheader("Agent Commands")
            # Fetch the available commands using the `Commands` class
            available_commands = agent_config["commands"]

            # Save the existing command state to prevent duplication
            existing_command_states = {
                command_name: command_status
                for command_name, command_status in available_commands.items()
            }

            all_commands_selected = st.checkbox("Select All Commands")

            for command_name, command_status in available_commands.items():
                if all_commands_selected:
                    available_commands[command_name] = True
                else:
                    toggle_status = st.checkbox(
                        command_name,
                        value=command_status,
                        key=command_name,
                    )
                    available_commands[command_name] = toggle_status
            if st.form_submit_button("Update Agent Commands"):
                ApiClient.update_agent_commands(
                    agent_name=agent_name, commands=available_commands
                )
    except Exception as e:
        st.error(f"Error loading agent configuration: {str(e)}")

if not new_agent:
    # Trigger actions on form submit
    if update_agent_settings_button:
        if agent_name:
            try:
                ApiClient.update_agent_commands(
                    agent_name=agent_name, commands=available_commands
                )
                ApiClient.update_agent_settings(
                    agent_name=agent_name, settings=agent_settings
                )
                st.success(f"Agent '{agent_name}' updated.")
            except Exception as e:
                st.error(f"Error updating agent: {str(e)}")

    if wipe_memories_button:
        if agent_name:
            try:
                ApiClient.wipe_agent_memories(agent_name=agent_name)
                st.success(f"Memories of agent '{agent_name}' wiped.")
            except Exception as e:
                st.error(f"Error wiping agent's memories: {str(e)}")

    if delete_agent_button:
        if agent_name:
            try:
                ApiClient.delete_agent(agent_name=agent_name)
                st.success(f"Agent '{agent_name}' deleted.")
                st.session_state["new_agent_name"] = ""  # Reset the selected agent
                st.experimental_rerun()  # Rerun the app to update the agent list
            except Exception as e:
                st.error(f"Error deleting agent: {str(e)}")
        else:
            st.error("Agent name is required.")



import re
import os
import regex
import json
import time
import spacy
from datetime import datetime
from Agent import Agent
from Prompts import Prompts
from extensions.searxng import searxng
from Chain import Chain, get_chain_responses_file_path, create_command_suggestion_chain
from urllib.parse import urlparse
import logging
from concurrent.futures import Future
from playwright.async_api import async_playwright
from bs4 import BeautifulSoup


class Interactions:
    def __init__(self, agent_name: str = ""):
        if agent_name != "":
            self.agent_name = agent_name
            self.agent = Agent(self.agent_name)
            self.agent_commands = self.agent.get_commands_string()
            self.memories = self.agent.get_memories()
        else:
            self.agent_name = ""
            self.agent = None
            self.agent_commands = ""
            self.memories = None
        self.stop_running_event = None
        self.browsed_links = []
        self.failures = 0
        self.nlp = None

    def load_spacy_model(self):
        if not self.nlp:
            try:
                self.nlp = spacy.load("en_core_web_sm")
            except:
                spacy.cli.download("en_core_web_sm")
                self.nlp = spacy.load("en_core_web_sm")
        self.nlp.max_length = 99999999999999999999999

    def custom_format(self, string, **kwargs):
        if isinstance(string, list):
            string = "".join(str(x) for x in string)

        def replace(match):
            key = match.group(1)
            value = kwargs.get(key, match.group(0))
            if isinstance(value, list):
                return "".join(str(x) for x in value)
            else:
                return str(value)

        pattern = r"(?<!{){([^{}\n]+)}(?!})"
        result = re.sub(pattern, replace, string)
        return result

    def get_step_response(self, chain_name, step_number):
        base_path = os.path.join(os.getcwd(), "chains")
        file_path = os.path.normpath(
            os.path.join(base_path, chain_name, "responses.json")
        )
        if not file_path.startswith(base_path):
            raise ValueError("Invalid path, chain name must not contain slashes.")
        try:
            with open(file_path, "r") as f:
                responses = json.load(f)
            return responses.get(str(step_number))
        except:
            return ""

    def get_step_content(self, chain_name, prompt_content, user_input, agent_name):
        new_prompt_content = {}
        if isinstance(prompt_content, dict):
            for arg, value in prompt_content.items():
                if isinstance(value, str):
                    if "{user_input}" in value:
                        value = value.replace("{user_input}", user_input)
                    if "{agent_name}" in value:
                        value = value.replace("{agent_name}", agent_name)
                    if "{STEP" in value:
                        # Count how many times {STEP is in the value
                        step_count = value.count("{STEP")
                        for i in range(step_count):
                            # Get the step number from value between {STEP and }
                            new_step_number = int(value.split("{STEP")[1].split("}")[0])
                            # get the response from the step number
                            step_response = self.get_step_response(
                                chain_name=chain_name, step_number=new_step_number
                            )
                            # replace the {STEPx} with the response
                            if step_response:
                                resp = (
                                    step_response["response"]
                                    if "response" in step_response
                                    else f"{step_response}"
                                )
                                value = value.replace(
                                    f"{{STEP{new_step_number}}}",
                                    f"{resp}",
                                )
                new_prompt_content[arg] = value
        elif isinstance(prompt_content, str):
            new_prompt_content = prompt_content
            if "{user_input}" in prompt_content:
                new_prompt_content = new_prompt_content.replace(
                    "{user_input}", user_input
                )
            if "{agent_name}" in new_prompt_content:
                new_prompt_content = new_prompt_content.replace(
                    "{agent_name}", agent_name
                )
            if "{STEP" in prompt_content:
                step_count = value.count("{STEP")
                for i in range(step_count):
                    # Get the step number from value between {STEP and }
                    new_step_number = int(
                        prompt_content.split("{STEP")[1].split("}")[0]
                    )
                    # get the response from the step number
                    step_response = self.get_step_response(
                        chain_name=chain_name, step_number=new_step_number
                    )
                    # replace the {STEPx} with the response
                    if step_response:
                        resp = (
                            step_response["response"]
                            if "response" in step_response
                            else f"{step_response}"
                        )
                        new_prompt_content = prompt_content.replace(
                            f"{{STEP{new_step_number}}}", f"{resp}"
                        )
            if new_prompt_content == {}:
                new_prompt_content = prompt_content
        return new_prompt_content

    async def format_prompt(
        self,
        user_input: str = "",
        top_results: int = 5,
        prompt="",
        chain_name="",
        step_number=0,
        memories=None,
        **kwargs,
    ):
        cp = Prompts()
        if prompt == "":
            prompt = user_input
        else:
            try:
                prompt = cp.get_prompt(prompt_name=prompt, model=self.agent.AI_MODEL)
            except:
                prompt = prompt
        if top_results == 0:
            context = "None"
        else:
            try:
                context = await memories.context_agent(
                    query=user_input, top_results_num=top_results
                )
            except:
                context = "None."
        command_list = self.agent.get_commands_string()
        if chain_name != "":
            try:
                for arg, value in kwargs.items():
                    if "{STEP" in value:
                        # get the response from the step number
                        step_response = self.get_step_response(
                            chain_name=chain_name, step_number=step_number
                        )
                        # replace the {STEPx} with the response
                        value = value.replace(f"{{STEP{step_number}}}", step_response)
                        kwargs[arg] = value
            except:
                logging.info("No args to replace.")
            if "{STEP" in prompt:
                step_response = self.get_step_response(
                    chain_name=chain_name, step_number=step_number
                )
                prompt = prompt.replace(f"{{STEP{step_number}}}", step_response)
            if "{STEP" in user_input:
                step_response = self.get_step_response(
                    chain_name=chain_name, step_number=step_number
                )
                user_input = user_input.replace(f"{{STEP{step_number}}}", step_response)
        try:
            working_directory = self.agent.AGENT_CONFIG["settings"]["WORKING_DIRECTORY"]
        except:
            working_directory = "./WORKSPACE"

        formatted_prompt = self.custom_format(
            string=prompt,
            user_input=user_input,
            agent_name=self.agent_name,
            COMMANDS=self.agent_commands,
            context=context,
            command_list=command_list,
            date=datetime.now().strftime("%B %d, %Y %I:%M %p"),
            working_directory=working_directory,
            **kwargs,
        )

        if not self.nlp:
            self.load_spacy_model()
        tokens = len(self.nlp(formatted_prompt))
        logging.info(f"FORMATTED PROMPT: {formatted_prompt}")
        return formatted_prompt, prompt, tokens

    async def run(
        self,
        user_input: str = "",
        prompt: str = "",
        context_results: int = 5,
        websearch: bool = False,
        websearch_depth: int = 3,
        learn_file: str = "",
        chain_name: str = "",
        step_number: int = 0,
        shots: int = 1,
        disable_memory: bool = False,
        **kwargs,
    ):
        shots = int(shots)
        if learn_file != "":
            try:
                learning_file = await self.memories.mem_read_file(file_path=learn_file)
            except:
                return "Failed to read file."
            if learning_file == False:
                return "Failed to read file."
        formatted_prompt, unformatted_prompt, tokens = await self.format_prompt(
            user_input=user_input,
            top_results=context_results,
            prompt=prompt,
            chain_name=chain_name,
            step_number=step_number,
            memories=self.memories,
            **kwargs,
        )
        if websearch:
            if user_input == "":
                if "primary_objective" in kwargs and "task" in kwargs:
                    search_string = f"Primary Objective: {kwargs['primary_objective']}\n\nTask: {kwargs['task']}"
            else:
                search_string = user_input
            if search_string != "":
                await self.websearch_agent(
                    user_input=search_string, depth=websearch_depth
                )
        try:
            # Workaround for non-threaded providers
            run_response = await self.agent.instruct(formatted_prompt, tokens=tokens)
            self.response = (
                run_response.result()
                if isinstance(run_response, Future)
                else run_response
            )
        except Exception as e:
            logging.info(f"Error: {e}")
            logging.info(f"PROMPT CONTENT: {formatted_prompt}")
            logging.info(f"TOKENS: {tokens}")
            self.failures += 1
            if self.failures == 5:
                self.failures == 0
                logging.info("Failed to get a response 5 times in a row.")
                return None
            logging.info(f"Retrying in 10 seconds...")
            time.sleep(10)
            if context_results > 0:
                context_results = context_results - 1
            self.response = await self.run(
                user_input=user_input,
                prompt=prompt,
                context_results=context_results,
                **kwargs,
            )

        # Handle commands if the prompt contains the {COMMANDS} placeholder
        # We handle command injection that DOESN'T allow command execution by using {command_list} in the prompt
        if "{COMMANDS}" in unformatted_prompt:
            execution_response = await self.execution_agent(
                execution_response=self.response,
                user_input=user_input,
                context_results=context_results,
                **kwargs,
            )
            return_response = ""
            if bool(self.agent.AUTONOMOUS_EXECUTION) == True:
                try:
                    self.response = json.loads(self.response)
                    if "response" in self.response:
                        return_response = self.response["response"]
                    if "commands" in self.response:
                        if self.response["commands"] != {}:
                            return_response += (
                                f"\n\nCommands Executed:\n{self.response['commands']}"
                            )
                    if execution_response:
                        return_response += (
                            f"\n\nCommand Execution Response:\n{execution_response}"
                        )
                except:
                    return_response = self.response
            else:
                return_response = f"{self.response}\n\n{execution_response}"
            self.response = return_response
        logging.info(f"Response: {self.response}")
        if self.response != "" and self.response != None:
            if disable_memory == False:
                try:
                    await self.memories.store_result(
                        input=user_input, result=self.response
                    )
                except:
                    pass
            if prompt == "Chat":
                self.agent.log_interaction(role="USER", message=user_input)
            else:
                self.agent.log_interaction(role="USER", message=formatted_prompt)
            self.agent.log_interaction(role=self.agent_name, message=self.response)

        if shots > 1:
            responses = [self.response]
            for shot in range(shots - 1):
                shot_response = await self.run(
                    user_input=user_input,
                    prompt=prompt,
                    context_results=context_results,
                    shots=shots - 1,
                    chain_name=chain_name,
                    step_number=step_number,
                    **kwargs,
                )
                time.sleep(1)
                responses.append(shot_response)
            return "\n".join(
                [
                    f"Response {shot + 1}:\n{response}"
                    for shot, response in enumerate(responses)
                ]
            )
        return self.response

    async def run_chain_step(
        self, step: dict = {}, chain_name="", user_input="", agent_override=""
    ):
        if step:
            if "prompt_type" in step:
                if agent_override != "":
                    self.agent_name = agent_override
                else:
                    self.agent_name = step["agent_name"]
                self.agent = Agent(self.agent_name)
                self.agent_commands = self.agent.get_commands_string()
                self.memories = self.agent.get_memories()
                prompt_type = step["prompt_type"]
                step_number = step["step"]
                if "prompt_name" in step["prompt"]:
                    prompt_name = step["prompt"]["prompt_name"]
                else:
                    prompt_name = ""
                args = self.get_step_content(
                    chain_name=chain_name,
                    prompt_content=step["prompt"],
                    user_input=user_input,
                    agent_name=self.agent_name,
                )
                if prompt_type == "Command":
                    return await self.agent.execute(
                        command_name=args["command_name"],
                        command_args=args,
                    )
                elif prompt_type == "Prompt":
                    result = await self.run(
                        user_input=user_input,
                        prompt=prompt_name,
                        chain_name=chain_name,
                        step_number=step_number,
                        **args,
                    )
                elif prompt_type == "Chain":
                    result = await self.run_chain(
                        chain_name=args["chain"],
                        user_input=args["input"],
                        agent_override=self.agent_name,
                        all_responses=False,
                    )
        if result:
            return result
        else:
            return None

    async def run_chain(
        self,
        chain_name,
        user_input=None,
        all_responses=True,
        agent_override="",
        from_step=1,
    ):
        chain = Chain()
        file_path = get_chain_responses_file_path(chain_name=chain_name)
        chain_data = chain.get_chain(chain_name=chain_name)
        if chain_data == {}:
            return f"Chain `{chain_name}` not found."
        logging.info(f"Running chain '{chain_name}'")
        responses = {}  # Create a dictionary to hold responses.
        last_response = ""
        for step_data in chain_data["steps"]:
            if int(step_data["step"]) >= int(from_step):
                if "prompt" in step_data and "step" in step_data:
                    step = {}
                    step["agent_name"] = (
                        agent_override
                        if agent_override != ""
                        else step_data["agent_name"]
                    )
                    step["prompt_type"] = step_data["prompt_type"]
                    step["prompt"] = step_data["prompt"]
                    logging.info(
                        f"Running step {step_data['step']} with agent {step['agent_name']}."
                    )
                    step_response = await self.run_chain_step(
                        step=step_data,
                        chain_name=chain_name,
                        user_input=user_input,
                        agent_override=agent_override,
                    )  # Get the response of the current step.
                    step["response"] = step_response
                    last_response = step_response
                    responses[step_data["step"]] = step  # Store the response.
                    logging.info(f"Response: {step_response}")
                    # Write the responses to the json file.
                    dt = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                    with open(file_path, "w") as f:
                        json.dump(responses, f)
        if all_responses == True:
            return responses
        else:
            # Return only the last response in the chain.
            return last_response

    async def smart_instruct(
        self,
        user_input: str = "Write a tweet about AI.",
        shots: int = 3,
        learn_file: str = "",
        objective: str = None,
        **kwargs,
    ):
        answers = []
        # Do multi shots of prompt to get N different answers to be validated
        answers.append(
            await self.run(
                user_input=user_input,
                prompt="SmartInstruct-StepByStep"
                if objective == None
                else "SmartTask-StepByStep",
                context_results=6,
                websearch=True,
                websearch_depth=3,
                shots=shots,
                learn_file=learn_file,
                objective=objective,
                **kwargs,
            )
        )
        if shots > 1:
            for i in range(shots - 1):
                answers.append(
                    await self.run(
                        user_input=user_input,
                        prompt="SmartInstruct-StepByStep"
                        if objective == None
                        else "SmartTask-StepByStep",
                        context_results=6,
                        shots=shots,
                        objective=objective,
                        **kwargs,
                    )
                )
        answer_str = ""
        for i, answer in enumerate(answers):
            answer_str += f"Answer {i + 1}:\n{answer}\n\n"
        researcher = await self.run(
            user_input=answer_str,
            prompt="SmartInstruct-Researcher",
            shots=shots,
            **kwargs,
        )
        resolver = await self.run(
            user_input=researcher,
            prompt="SmartInstruct-Resolver",
            shots=shots,
            **kwargs,
        )
        execution_response = await self.run(
            user_input=f"{user_input}\nContext:\n{resolver}",
            prompt="instruct",
            **kwargs,
        )
        response = f"{resolver}\n\n{execution_response}"
        return response

    async def smart_chat(
        self,
        user_input: str = "Write a tweet about AI.",
        shots: int = 3,
        learn_file: str = "",
        **kwargs,
    ):
        answers = []
        answers.append(
            await self.run(
                user_input=user_input,
                prompt="SmartChat-StepByStep",
                context_results=6,
                websearch=True,
                websearch_depth=3,
                shots=shots,
                learn_file=learn_file,
                **kwargs,
            )
        )
        # Do multi shots of prompt to get N different answers to be validated
        if shots > 1:
            for i in range(shots - 1):
                answers.append(
                    await self.run(
                        user_input=user_input,
                        prompt="SmartChat-StepByStep",
                        context_results=6,
                        shots=shots,
                        **kwargs,
                    )
                )
        answer_str = ""
        for i, answer in enumerate(answers):
            answer_str += f"Answer {i + 1}:\n{answer}\n\n"
        researcher = await self.run(
            user_input=answer_str,
            prompt="SmartChat-Researcher",
            context_results=6,
            shots=shots,
            **kwargs,
        )
        resolver = await self.run(
            user_input=researcher,
            prompt="SmartChat-Resolver",
            context_results=6,
            shots=shots,
            **kwargs,
        )
        return resolver

    # Worker Sub-Agents
    async def validation_agent(
        self, user_input, execution_response, context_results, **kwargs
    ):
        try:
            pattern = regex.compile(r"\{(?:[^{}]|(?R))*\}")
            cleaned_json = pattern.findall(execution_response)
            if len(cleaned_json) == 0:
                return {}
            if isinstance(cleaned_json, list):
                cleaned_json = cleaned_json[0]
            response = json.loads(cleaned_json)
            return response
        except:
            logging.info("INVALID JSON RESPONSE")
            logging.info(execution_response)
            logging.info("... Trying again.")
            if context_results != 0:
                context_results = context_results - 1
            else:
                context_results = 0
            execution_response = await self.run(
                user_input=user_input, context_results=context_results, **kwargs
            )
            return await self.validation_agent(
                user_input=user_input,
                execution_response=execution_response,
                context_results=context_results,
                **kwargs,
            )

    async def execution_agent(
        self, execution_response, user_input, context_results, **kwargs
    ):
        validated_response = await self.validation_agent(
            user_input=user_input,
            execution_response=execution_response,
            context_results=context_results,
            **kwargs,
        )
        if "commands" in validated_response:
            for command_name, command_args in validated_response["commands"].items():
                # Search for the command in the available_commands list, and if found, use the command's name attribute for execution
                if command_name is not None:
                    for available_command in self.agent.available_commands:
                        if command_name == available_command["friendly_name"]:
                            # Check if the command is a valid command in the self.avent.available_commands list
                            try:
                                if bool(self.agent.AUTONOMOUS_EXECUTION) == True:
                                    command_output = await self.agent.execute(
                                        command_name=command_name,
                                        command_args=command_args,
                                    )
                                else:
                                    command_output = create_command_suggestion_chain(
                                        agent_name=self.agent_name,
                                        command_name=command_name,
                                        command_args=command_args,
                                    )
                            except Exception as e:
                                logging.info("Command validation failed, retrying...")
                                validate_command = await self.run(
                                    user_input=user_input,
                                    prompt="ValidationFailed",
                                    command_name=command_name,
                                    command_args=command_args,
                                    command_output=e,
                                    context_results=context_results,
                                    **kwargs,
                                )
                                return await self.execution_agent(
                                    execution_response=validate_command,
                                    user_input=user_input,
                                    context_results=context_results,
                                    **kwargs,
                                )
                            logging.info(
                                f"Command {command_name} executed successfully with args {command_args}. Command Output: {command_output}"
                            )
                            response = f"\nExecuted Command:{command_name} with args {command_args}.\nCommand Output: {command_output}\n"
                            return response
                else:
                    if command_name == "None.":
                        return "\nNo commands were executed.\n"
                    else:
                        return f"\Command not recognized: `{command_name}`."
        else:
            return "\nNo commands were executed.\n"

    async def get_web_content(self, url):
        try:
            async with async_playwright() as p:
                browser = await p.chromium.launch()
                context = await browser.new_context()
                page = await context.new_page()
                await page.goto(url)
                content = await page.content()

                # Scrape links and their titles
                links = await page.query_selector_all("a")
                link_list = []
                for link in links:
                    title = await page.evaluate("(link) => link.textContent", link)
                    href = await page.evaluate("(link) => link.href", link)
                    link_list.append((title, href))

                await browser.close()
                soup = BeautifulSoup(content, "html.parser")
                text_content = soup.get_text()
                text_content = " ".join(text_content.split())
                return text_content, link_list
        except:
            return None, None

    async def resursive_browsing(self, user_input, links):
        chunk_size = int(int(self.agent.MAX_TOKENS) / 2)
        try:
            words = links.split()
            links = [
                word for word in words if urlparse(word).scheme in ["http", "https"]
            ]
        except:
            links = links
        if links is not None:
            for link in links:
                if "href" in link:
                    try:
                        url = link["href"]
                    except:
                        url = link
                else:
                    url = link
                url = re.sub(r"^.*?(http)", r"http", url)
                # Check if url is an actual url
                if url.startswith("http"):
                    logging.info(f"Scraping: {url}")
                    if url not in self.browsed_links:
                        self.browsed_links.append(url)
                        (
                            collected_data,
                            link_list,
                        ) = await self.get_web_content(url=url)
                        # Split the collected data into agent max tokens / 2 character chunks
                        if collected_data is not None:
                            if len(collected_data) > 0:
                                chunks = [
                                    collected_data[i : i + chunk_size]
                                    for i in range(
                                        0,
                                        len(collected_data),
                                        chunk_size,
                                    )
                                ]
                                for chunk in chunks:
                                    summarized_content = await self.run(
                                        user_input=user_input,
                                        prompt="Summarize Web Content",
                                        link=url,
                                        chunk=chunk,
                                        disable_memory=True,
                                    )
                                    if not summarized_content.startswith("None"):
                                        try:
                                            await self.memories.store_result(
                                                input=user_input,
                                                result=summarized_content,
                                                external_source_name=url,
                                            )
                                        except:
                                            logging.info(
                                                f"Failed to store result for {url}. Moving on..."
                                            )
                        if link_list is not None:
                            if len(link_list) > 0:
                                if len(link_list) > 5:
                                    link_list = link_list[:3]
                                try:
                                    pick_a_link = await self.run(
                                        user_input=user_input,
                                        prompt="Pick-a-Link",
                                        links=link_list,
                                    )
                                    if not pick_a_link.startswith("None"):
                                        await self.resursive_browsing(
                                            user_input=user_input, links=pick_a_link
                                        )
                                except:
                                    logging.info(f"Issues reading {url}. Moving on...")

    async def websearch_agent(
        self,
        user_input: str = "What are the latest breakthroughs in AI?",
        depth: int = 3,
    ):
        results = await self.run(
            user_input=user_input, prompt="WebSearch", disable_memory=True
        )
        results = results.split("\n")
        for result in results:
            search_string = result.lstrip("0123456789. ")
            try:
                searx_server = self.agent.PROVIDER_SETTINGS["SEARXNG_INSTANCE_URL"]
            except:
                searx_server = ""
            try:
                links = await searxng(SEARXNG_INSTANCE_URL=searx_server).search(
                    query=search_string
                )
                if len(links) > depth:
                    links = links[:depth]
            except:
                links = None
            if links is not None:
                await self.resursive_browsing(user_input=user_input, links=links)



import os


def get_prompt_file_path(prompt_name, model="default"):
    base_path = os.path.join(os.getcwd(), "prompts")
    base_model_path = os.path.normpath(os.path.join(os.getcwd(), "prompts", model))
    model_prompt_file = os.path.normpath(
        os.path.join(base_model_path, f"{prompt_name}.txt")
    )
    default_prompt_file = os.path.normpath(
        os.path.join(base_path, f"{prompt_name}.txt")
    )
    if (
        not base_model_path.startswith(base_path)
        or not model_prompt_file.startswith(base_model_path)
        or not default_prompt_file.startswith(base_path)
    ):
        raise ValueError(
            "Invalid file path. Prompt name cannot contain '/', '\\' or '..' in"
        )
    if not os.path.exists(base_path):
        os.mkdir(base_path)
    if not os.path.exists(base_model_path):
        os.mkdir(base_model_path)
    prompt_file = (
        model_prompt_file if os.path.isfile(model_prompt_file) else default_prompt_file
    )
    return prompt_file


class Prompts:
    def add_prompt(self, prompt_name, prompt):
        # if prompts folder does not exist, create it
        file_path = get_prompt_file_path(prompt_name=prompt_name)
        # if prompt file does not exist, create it
        if not os.path.exists(file_path):
            with open(file_path, "w") as f:
                f.write(prompt)

    def get_prompt(self, prompt_name, model="default"):
        prompt_file = get_prompt_file_path(prompt_name=prompt_name, model=model)
        with open(prompt_file, "r") as f:
            prompt = f.read()
            return prompt

    def get_prompts(self):
        # Get all files in prompts folder that end in .txt and replace .txt with empty string
        prompts = []
        for file in os.listdir("prompts"):
            if file.endswith(".txt"):
                prompts.append(file.replace(".txt", ""))
        return prompts

    def get_prompt_args(self, prompt_name):
        prompt = self.get_prompt(prompt_name=prompt_name)
        # Find anything in the file between { and } and add them to a list to return
        prompt_vars = []
        for word in prompt.split():
            if word.startswith("{") and word.endswith("}"):
                prompt_vars.append(word[1:-1])
        return prompt_vars

    def delete_prompt(self, prompt_name):
        prompt_file = get_prompt_file_path(prompt_name=prompt_name)
        os.remove(prompt_file)

    def update_prompt(self, prompt_name, prompt):
        prompt_file = get_prompt_file_path(prompt_name=prompt_name)
        with open(prompt_file, "w") as f:
            f.write(prompt)

    def get_model_prompt(self, prompt_name, model="default"):
        prompt_file = get_prompt_file_path(prompt_name=prompt_name, model=model)
        with open(prompt_file, "r") as f:
            return f.read()



import streamlit as st


def agixt_docs():
    st.markdown(
        "[![Documentation](https://img.shields.io/badge/Documentation-AGiXT-blue?logo=github&style=plastic)](https://josh-xt.github.io/AGiXT/) [![GitHub](https://img.shields.io/badge/GitHub-AGiXT-blue?logo=github&style=plastic)](https://github.com/Josh-XT/AGiXT) [![Discord](https://img.shields.io/discord/1097720481970397356?label=Discord&logo=discord&logoColor=white&style=plastic&color=5865f2)](https://discord.gg/d3TkHRZcjD) [![Twitter](https://img.shields.io/badge/Twitter-Follow_@Josh__XT-blue?logo=twitter&style=plastic)](https://twitter.com/Josh_XT)"
    )



import requests
import time
import logging


# Custom OpenAI Style Provider
class CustomProvider:
    def __init__(
        self,
        API_KEY: str = "",
        API_URI: str = "https://api.openai.com/v1/engines/davinci/completions",
        AI_MODEL: str = "gpt-3.5-turbo-16k-0613",
        AI_TEMPERATURE: float = 0.7,
        AI_TOP_P: float = 0.7,
        MAX_TOKENS: int = 4096,
        WAIT_BETWEEN_REQUESTS: int = 0,
        WAIT_AFTER_FAILURE: int = 3,
        **kwargs,
    ):
        self.requirements = ["requests"]
        self.AI_MODEL = AI_MODEL if AI_MODEL else "gpt-3.5-turbo-16k-0613"
        self.AI_TEMPERATURE = AI_TEMPERATURE if AI_TEMPERATURE else 0.7
        self.AI_TOP_P = AI_TOP_P if AI_TOP_P else 0.7
        self.MAX_TOKENS = MAX_TOKENS if MAX_TOKENS else 16000
        self.API_KEY = API_KEY
        self.API_URI = API_URI
        self.WAIT_AFTER_FAILURE = WAIT_AFTER_FAILURE if WAIT_AFTER_FAILURE else 3
        self.WAIT_BETWEEN_REQUESTS = (
            WAIT_BETWEEN_REQUESTS if WAIT_BETWEEN_REQUESTS else 0
        )

    async def instruct(self, prompt, tokens: int = 0):
        if int(self.WAIT_BETWEEN_REQUESTS) > 0:
            time.sleep(int(self.WAIT_BETWEEN_REQUESTS))
        max_new_tokens = int(self.MAX_TOKENS) - tokens
        if not self.AI_MODEL.startswith("gpt-"):
            # Use completion API
            params = {
                "prompt": prompt,
                "model": self.AI_MODEL,
                "temperature": float(self.AI_TEMPERATURE),
                "max_tokens": max_new_tokens,
                "top_p": float(self.AI_TOP_P),
                "frequency_penalty": 0,
                "presence_penalty": 0,
                "stream": False,
            }
        else:
            # Use chat completion API
            params = {
                "messages": [{"role": "user", "content": prompt}],
                "model": self.AI_MODEL,
                "temperature": float(self.AI_TEMPERATURE),
                "max_tokens": max_new_tokens,
                "top_p": float(self.AI_TOP_P),
                "stream": False,
            }

        response = requests.post(
            self.API_URI,
            headers={"Authorization": f"Bearer {self.API_KEY}"},
            json=params,
        )
        data = response.json()
        if data:
            if "choices" in data:
                if data["choices"]:
                    if "text" in data["choices"][0]:
                        return data["choices"][0]["text"].strip()
                    if "message" in data["choices"][0]:
                        return data["choices"][0]["message"]["content"].strip()
            if "error" in data:
                logging.info(f"Custom API Error: {data}")
                if int(self.WAIT_AFTER_FAILURE) > 0:
                    time.sleep(int(self.WAIT_AFTER_FAILURE))
                    return await self.instruct(prompt=prompt, tokens=tokens)
            return str(data)



import requests
import random


class LlamacppapiProvider:
    def __init__(
        self,
        AI_PROVIDER_URI: str = "http://localhost:8000",
        MAX_TOKENS: int = 2048,
        AI_TEMPERATURE: float = 0.7,
        AI_MODEL: str = "default",
        STOP_SEQUENCE: str = "</s>",
        **kwargs,
    ):
        self.AI_PROVIDER_URI = (
            AI_PROVIDER_URI if AI_PROVIDER_URI else "http://localhost:8000"
        )
        self.AI_TEMPERATURE = AI_TEMPERATURE if AI_TEMPERATURE else 0.7
        self.MAX_TOKENS = MAX_TOKENS if MAX_TOKENS else 2048
        self.AI_MODEL = AI_MODEL if AI_MODEL else "default"
        self.STOP_SEQUENCE = STOP_SEQUENCE if STOP_SEQUENCE else "</s>"
        self.MAX_TOKENS = int(self.MAX_TOKENS) if self.MAX_TOKENS else 2048

    async def instruct(self, prompt, tokens: int = 0):
        params = {
            "prompt": prompt,
            "temperature": float(self.AI_TEMPERATURE),
            "stop": self.STOP_SEQUENCE,
            "seed": random.randint(1, 1000000000),
        }
        response = requests.post(f"{self.AI_PROVIDER_URI}/v1/completions", json=params)
        data = response.json()
        print(data)
        choices = data["choices"]
        if choices:
            return choices[0]["text"]
        return None



import os

import git
from github import Github
from Extensions import Extensions


class github(Extensions):
    def __init__(
        self,
        GITHUB_USERNAME: str = "",
        GITHUB_API_KEY: str = "",
        WORKING_DIRECTORY: str = "./WORKSPACE",
        **kwargs,
    ):
        self.GITHUB_USERNAME = GITHUB_USERNAME
        self.GITHUB_API_KEY = GITHUB_API_KEY
        self.WORKING_DIRECTORY = WORKING_DIRECTORY
        self.commands = {"Clone Github Repository": self.clone_repo}
        if self.GITHUB_USERNAME and self.GITHUB_API_KEY:
            self.commands["Create Github Repository"] = self.create_repo

    async def clone_repo(self, repo_url: str) -> str:
        split_url = repo_url.split("//")
        if self.GITHUB_USERNAME is not None and self.GITHUB_API_KEY is not None:
            auth_repo_url = f"//{self.GITHUB_USERNAME}:{self.GITHUB_API_KEY}@".join(
                split_url
            )
        else:
            auth_repo_url = "//".join(split_url)
        try:
            repo_name = repo_url.split("/")[-1]
            repo_dir = os.path.join(self.WORKING_DIRECTORY, repo_name)
            if os.path.exists(repo_dir):
                return f"""{repo_dir} already exists"""
            git.Repo.clone_from(
                url=auth_repo_url,
                to_path=repo_dir,
            )
            return f"Cloned {repo_url} to {repo_dir}"
        except Exception as e:
            return f"Error: {str(e)}"

    async def create_repo(self, repo_name: str, content_of_readme: str) -> str:
        g = Github(self.GITHUB_API_KEY)
        user = g.get_user(self.GITHUB_USERNAME)
        repo = user.create_repo(repo_name, private=True)
        repo_url = repo.clone_url
        repo_dir = f"./{repo_name}"
        repo = git.Repo.init(repo_dir)
        with open(f"{repo_dir}/README.md", "w") as f:
            f.write(content_of_readme)
        repo.git.add(A=True)
        repo.git.commit(m="Added README")
        repo.create_remote("origin", repo_url)
        repo.git.push("origin", "HEAD:main")
        return repo_url



from ApiClient import ApiClient
import streamlit as st
import json
from components.selectors import command_selection, prompt_selection, chain_selection


def add_new_step(chain_name, step_number, agents):
    st.markdown(f"## Add Chain Step {step_number}")
    agent_name = st.selectbox(
        "Select Agent",
        options=[""] + [agent["name"] for agent in agents],
        index=0,
        key="add_step_agent_name",
    )
    prompt_type = st.selectbox(
        "Select Step Type",
        [""] + ["Command", "Prompt", "Chain"],
        key="add_step_prompt_type",
    )

    if prompt_type == "Command":
        prompt = command_selection(step_number=step_number)
    elif prompt_type == "Prompt":
        prompt = prompt_selection(step_number=step_number)
    elif prompt_type == "Chain":
        prompt = chain_selection(step_number=step_number)
    else:
        prompt = {}
    if st.button("Add New Step", key=f"add_step_{step_number}"):
        if chain_name and step_number and agent_name and prompt_type and prompt:
            ApiClient.add_step(
                chain_name=chain_name,
                step_number=step_number,
                agent_name=agent_name,
                prompt_type=prompt_type,
                prompt=prompt,
            )
            st.success(f"Step added to chain '{chain_name}'.")
            st.experimental_rerun()
        else:
            st.error("All fields are required.")


def modify_step(chain_name, step, agents):
    step_number = step["step"]
    agent_name = step["agent_name"]
    prompt_type = step["prompt_type"]
    step_prompt = step["prompt"]
    modify_step_number = st.number_input(
        "Step Number",
        min_value=1,
        step=1,
        value=step_number,
        key=f"step_num_{step_number}",
    )
    modify_agent_name = st.selectbox(
        "Select Agent",
        options=[""] + [agent["name"] for agent in agents],
        index=([agent["name"] for agent in agents].index(agent_name) + 1)
        if agent_name in [agent["name"] for agent in agents]
        else 0,
        key=f"agent_name_{step_number}",
    )
    modify_prompt_type = st.selectbox(
        "Select Step Type",
        options=["", "Command", "Prompt", "Chain"],
        index=["", "Command", "Prompt", "Chain"].index(prompt_type),
        key=f"prompt_type_{step_number}",
    )
    if prompt_type == "Command":
        prompt = command_selection(prompt=step_prompt, step_number=step_number)
    elif prompt_type == "Prompt":
        prompt = prompt_selection(prompt=step_prompt, step_number=step_number)
    elif prompt_type == "Chain":
        prompt = chain_selection(prompt=step_prompt, step_number=step_number)
    else:
        prompt = {}
    if st.button("Modify Step", key=f"modify_step_{step_number}"):
        if chain_name and step_number and agent_name and prompt_type and prompt:
            ApiClient.update_step(
                chain_name=chain_name,
                step_number=modify_step_number,
                agent_name=modify_agent_name,
                prompt_type=modify_prompt_type,
                prompt=prompt,
            )
            st.success(f"Step modified in chain '{chain_name}'.")
            st.experimental_rerun()
        else:
            st.error("All fields are required.")


def modify_chain(chain_name, agents):
    if chain_name:
        chain = ApiClient.get_chain(chain_name=chain_name)
        export_button = st.download_button(
            "Export Chain",
            data=json.dumps(chain, indent=4),
            file_name=f"{chain_name}.json",
            mime="application/json",
        )
        st.markdown(f"## Modifying Chain: {chain_name}")
        if chain:
            for step in chain["steps"]:
                if step is not None:
                    new_step = modify_step(
                        chain_name=chain_name,
                        step=step,
                        agents=agents,
                    )

        if len(chain["steps"]) > 0:
            step_number = max([s["step"] for s in chain["steps"]]) + 1 if chain else 1
        else:
            step_number = 1

        add_step = add_new_step(
            chain_name=chain_name,
            step_number=step_number,
            agents=agents,
        )
    else:
        st.warning("Please select a chain to manage steps.")



import socket, errno, subprocess, os, logging
import streamlit as st


def verify_backend():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    try:
        s.connect(("127.0.0.1", 7437))
        s.close()
    except socket.error as e:
        if e.errno == errno.EADDRINUSE:
            print("Port is already in use")
        else:
            try:
                subprocess.Popen(
                    os.system(
                        "cd ../agixt/ && uvicorn app:app --host 0.0.0.0 --port 7437 --workers 4 && cd ../streamlit/"
                    )
                )
                logging.info("LAUNCHED FOR YOU")
                s = None
                st.experimental_rerun()
            except Exception as e:
                logging.info(e)
                print("Press CTRL + C to exit.")
                s.close()

    if s != None:
        s.close()



from .Interactions import Interactions
from .Memories import Memories
from .Chain import Chain
from .Extensions import Extensions
from .Agent import Agent
from .Embedding import Embedding
from .Prompts import Prompts



from Chain import Chain
from Extensions import Extensions
from Interactions import Interactions
from Prompts import Prompts
import datetime
import json
import requests


class agixt_chain(Extensions):
    def __init__(self, **kwargs):
        self.chains = Chain().get_chains()
        self.commands = {
            "Create Task Chain": self.create_task_chain,
            "Generate Extension from OpenAPI": self.generate_openapi_chain,
        }
        if self.chains != None:
            for chain in self.chains:
                if "name" in chain:
                    self.commands.update(
                        {f"Run Chain: {chain['name']}": self.run_chain}
                    )

    async def create_task_chain(
        self,
        agent: str,
        primary_objective: str,
        numbered_list_of_tasks: str,
        short_chain_description: str,
        smart_chain: bool = False,
        researching: bool = False,
    ):
        now = datetime.datetime.now()
        timestamp = now.strftime("%Y-%m-%d_%H-%M-%S")
        task_list = numbered_list_of_tasks.split("\n")
        new_task_list = []
        current_task = ""
        for task in task_list:
            if task and task[0] in [
                str(i) for i in range(10)
            ]:  # Check for task starting with a digit (0-9)
                if current_task:  # If there's a current task, add it to the list
                    new_task_list.append(
                        current_task.lstrip("0123456789.")
                    )  # Strip leading digits and periods
                current_task = task  # Start a new current task
            else:
                current_task += (
                    "\n" + task
                )  # If the line doesn't start with a number, it's a subtask - add it to the current task

        if current_task:  # Add the last task if it exists
            if "\n\n" in current_task:
                current_task = current_task.split("\n\n")[0]
            new_task_list.append(
                current_task.lstrip("0123456789.")
            )  # Strip leading digits and periods

        task_list = new_task_list

        chain_name = f"AI Generated Task - {short_chain_description} - {timestamp}"
        chain = Chain()
        chain.add_chain(chain_name=chain_name)
        i = 1
        for task in task_list:
            chain.add_chain_step(
                chain_name=chain_name,
                agent_name=agent,
                step_number=i,
                prompt_type="Chain",
                prompt={
                    "chain": "Smart Prompt",
                    "input": f"Primary Objective to keep in mind while working on the task: {primary_objective} \nThe only task to complete to move towards the objective: {task}",
                },
            )
            i += 1
            if smart_chain:
                if researching:
                    step_chain = "Smart Instruct"
                else:
                    step_chain = "Smart Instruct - No Research"
                chain.add_chain_step(
                    chain_name=chain_name,
                    agent_name=agent,
                    step_number=i,
                    prompt_type="Chain",
                    prompt={
                        "chain": step_chain,
                        "input": "{STEP" + str(i - 1) + "}",
                    },
                )
            else:
                chain.add_chain_step(
                    chain_name=chain_name,
                    agent_name=agent,
                    step_number=i,
                    prompt_type="Prompt",
                    prompt={
                        "prompt_name": "Task Execution",
                        "introduction": "{STEP" + str(i - 1) + "}",
                        "websearch": researching,
                        "websearch_depth": 3,
                        "context_results": 5,
                    },
                )
            i += 1
        return chain_name

    async def run_chain(self, chain: str = "", input: str = ""):
        await Interactions(agent_name="").run_chain(chain_name=chain, user_input=input)
        return "Chain started successfully."

    def parse_openapi(self, data):
        endpoints = []
        schemas = data.get("components", {}).get(
            "schemas", {}
        )  # get the global schemas

        def resolve_schema(ref):
            # remove the '#/components/schemas/' part
            schema_name = ref.replace("#/components/schemas/", "")
            return schemas.get(schema_name, {})

        if "paths" in data:
            for path, path_info in data["paths"].items():
                for method, method_info in path_info.items():
                    endpoint_info = {
                        "endpoint": path,
                        "method": method.upper(),
                        "summary": method_info.get("summary", ""),
                        "parameters": [],
                        "responses": [],
                        "requestBody": {},
                    }
                    if "parameters" in method_info:
                        for param in method_info["parameters"]:
                            param_info = {
                                "name": param.get("name", ""),
                                "in": param.get("in", ""),
                                "description": param.get("description", ""),
                                "required": param.get("required", False),
                                "type": param.get("schema", {}).get("type", "")
                                if "schema" in param
                                else "",
                            }
                            endpoint_info["parameters"].append(param_info)
                    if "requestBody" in method_info:
                        request_body = method_info["requestBody"]
                        content = request_body.get("content", {})
                        for content_type, content_info in content.items():
                            if "$ref" in content_info.get("schema", {}):
                                # resolve the reference into the actual schema
                                content_info["schema"] = resolve_schema(
                                    content_info["schema"]["$ref"]
                                )
                        endpoint_info["requestBody"] = {
                            "description": request_body.get("description", ""),
                            "required": request_body.get("required", False),
                            "content": content,
                        }
                    if "responses" in method_info:
                        for response, response_info in method_info["responses"].items():
                            response_info = {
                                "code": response,
                                "description": response_info.get("description", ""),
                            }
                            endpoint_info["responses"].append(response_info)
                    endpoints.append(endpoint_info)
        return endpoints

    def get_auth_type(self, openapi_data):
        # The "components" section contains the security schemes
        if (
            "components" in openapi_data
            and "securitySchemes" in openapi_data["components"]
        ):
            security_schemes = openapi_data["components"]["securitySchemes"]

            # Iterate over all security schemes
            for scheme_name, scheme_details in security_schemes.items():
                # The "type" field specifies the type of the scheme
                if "type" in scheme_details and scheme_details["type"] == "http":
                    # The "scheme" field provides more specific information
                    if "scheme" in scheme_details:
                        return scheme_details["scheme"]
        return "basic"

    async def generate_openapi_chain(
        self, agent: str, extension_name: str, openapi_json_url: str
    ):
        # Experimental currently.
        openapi_str = requests.get(openapi_json_url).text
        openapi_data = json.loads(openapi_str)
        endpoints = self.parse_openapi(data=openapi_data)
        auth_type = self.get_auth_type(openapi_data=openapi_data)
        extension_name = extension_name.lower().replace(" ", "_")
        chain_name = f"OpenAPI to Python Chain - {extension_name}"
        chain = Chain()
        chain.add_chain(chain_name=chain_name)
        i = 0
        for endpoint in endpoints:
            i += 1
            chain.add_chain_step(
                chain_name=chain_name,
                agent_name=agent,
                step_number=i,
                prompt_type="Prompt",
                prompt={
                    "prompt_name": "Convert OpenAPI Endpoint",
                    "api_endpoint_info": f"{endpoint}",
                },
            )
            i += 1
            chain.add_chain_step(
                chain_name=chain_name,
                agent_name=agent,
                step_number=i,
                prompt_type="Command",
                prompt={
                    "command_name": "Get Python Code from Response",
                    "response": "{STEP" + str(i - 1) + "}",
                },
            )
            i += 1
            chain.add_chain_step(
                chain_name=chain_name,
                agent_name=agent,
                step_number=i,
                prompt_type="Command",
                prompt={
                    "command_name": "Indent String for Python Code",
                    "string": "{STEP" + str(i - 1) + "}",
                    "indents": 1,
                },
            )
            i += 1
            chain.add_chain_step(
                chain_name=chain_name,
                agent_name=agent,
                step_number=i,
                prompt_type="Command",
                prompt={
                    "command_name": "Append to File",
                    "filename": f"{extension_name}_functions.py",
                    "text": "\n\n{STEP" + str(i - 1) + "}",
                },
            )
        i += 1
        chain.add_chain_step(
            chain_name=chain_name,
            agent_name=agent,
            step_number=i,
            prompt_type="Command",
            prompt={
                "command_name": "Read File",
                "filename": f"{extension_name}_functions.py",
            },
        )
        i += 1
        chain.add_chain_step(
            chain_name=chain_name,
            agent_name=agent,
            step_number=i,
            prompt_type="Command",
            prompt={
                "command_name": "Generate Commands Dictionary",
                "python_file_content": "{STEP" + str(i - 1) + "}",
            },
        )
        new_extension = Prompts().get_prompt(prompt_name="New Extension Format")
        new_extension = new_extension.replace("{extension_name}", extension_name)
        new_extension = new_extension.replace("extension_commands", "STEP" + str(i))
        new_extension = new_extension.replace(
            "extension_functions", "STEP" + str(i - 1)
        )
        new_extension = new_extension.replace("{auth_type}", auth_type)
        i += 1
        chain.add_chain_step(
            chain_name=chain_name,
            agent_name=agent,
            step_number=i,
            prompt_type="Command",
            prompt={
                "command_name": "Write to File",
                "filename": f"{extension_name}.py",
                "text": new_extension,
            },
        )
        return chain_name



import uvicorn
from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
from Interactions import Interactions
from Agent import Agent, add_agent, delete_agent, rename_agent, get_agents
from Chain import Chain, import_chain
from Prompts import Prompts
from typing import Optional, Dict, List, Any
from provider import get_provider_options, get_providers
from Embedding import get_embedding_providers
from Extensions import Extensions
import os
import logging
import base64
import time
import string
import random

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "version"), encoding="utf-8") as f:
    version = f.read().strip()

logging.basicConfig(
    level=os.environ.get("LOGLEVEL", "INFO"),
    format="%(asctime)s | %(levelname)s | %(message)s",
)

app = FastAPI(
    title="AGiXT",
    description="AGiXT is an Artificial Intelligence Automation platform for creating and managing AI agents. Visit the GitHub repo for more information or to report issues. https://github.com/Josh-XT/AGiXT/",
    version=version,
    docs_url="/",
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


class AgentName(BaseModel):
    agent_name: str


class AgentNewName(BaseModel):
    new_name: str


class AgentPrompt(BaseModel):
    user_input: str
    prompt_name: str
    prompt_args: dict
    websearch: bool
    websearch_depth: int
    context_results: int


class Objective(BaseModel):
    objective: str


class Prompt(BaseModel):
    prompt: str


class PromptName(BaseModel):
    prompt_name: str


class PromptList(BaseModel):
    prompts: List[str]


class Completions(BaseModel):
    # Everything in this class except prompt, n, and model (agent_name) are unused currently.
    prompt: str
    max_tokens: int = 100
    temperature: float = 0.9
    top_p: float = 1.0
    n: int = 1
    stream: bool = False
    logprobs: int = None
    stop: List[str] = None
    presence_penalty: float = 0.0
    frequency_penalty: float = 0.0
    best_of: int = 1
    echo: bool = False
    user: str = None
    model: str = None  # Model is actually the agent_name
    stop_sequence: List[str] = None
    metadata: Dict[str, str] = None


class ChainNewName(BaseModel):
    new_name: str


class ChainName(BaseModel):
    chain_name: str


class ChainData(BaseModel):
    chain_name: str
    steps: Dict[str, Any]


class RunChain(BaseModel):
    prompt: str
    agent_override: Optional[str] = ""
    all_responses: Optional[bool] = False
    from_step: Optional[int] = 1


class StepInfo(BaseModel):
    step_number: int
    agent_name: str
    prompt_type: str
    prompt: dict


class RunChainResponse(BaseModel):
    response: str
    agent_name: str
    prompt: dict
    prompt_type: str


class ChainStep(BaseModel):
    step_number: int
    agent_name: str
    prompt_type: str
    prompt: dict


class ChainStepNewInfo(BaseModel):
    old_step_number: int
    new_step_number: int


class ResponseMessage(BaseModel):
    message: str


class UrlInput(BaseModel):
    url: str


class FileInput(BaseModel):
    file_name: str
    file_content: str


class TaskOutput(BaseModel):
    output: str
    message: Optional[str] = None


class ToggleCommandPayload(BaseModel):
    command_name: str
    enable: bool


class CustomPromptModel(BaseModel):
    prompt_name: str
    prompt: str


class AgentSettings(BaseModel):
    agent_name: str
    settings: Dict[str, Any]


class AgentConfig(BaseModel):
    agent_name: str
    settings: Dict[str, Any]
    commands: Dict[str, Any]


class AgentCommands(BaseModel):
    agent_name: str
    commands: Dict[str, Any]


@app.get("/api/provider", tags=["Provider"])
async def getproviders():
    providers = get_providers()
    return {"providers": providers}


@app.get("/api/provider/{provider_name}", tags=["Provider"])
async def get_provider_settings(provider_name: str):
    settings = get_provider_options(provider_name=provider_name)
    return {"settings": settings}


@app.get("/api/embedding_providers", tags=["Provider"])
async def get_embed_providers():
    providers = get_embedding_providers()
    return {"providers": providers}


@app.post("/api/agent", tags=["Agent"])
async def addagent(agent: AgentSettings) -> Dict[str, str]:
    return add_agent(agent_name=agent.agent_name, provider_settings=agent.settings)


@app.post("/api/agent/import", tags=["Agent"])
async def import_agent(agent: AgentConfig) -> Dict[str, str]:
    return add_agent(
        agent_name=agent.agent_name,
        provider_settings=agent.settings,
        commands=agent.commands,
    )


@app.patch("/api/agent/{agent_name}", tags=["Agent"])
async def renameagent(agent_name: str, new_name: AgentNewName) -> ResponseMessage:
    rename_agent(agent_name=agent_name, new_name=new_name.new_name)
    return ResponseMessage(message="Agent renamed.")


@app.put("/api/agent/{agent_name}", tags=["Agent"])
async def update_agent_settings(
    agent_name: str, settings: AgentSettings
) -> ResponseMessage:
    update_config = Agent(agent_name=agent_name).update_agent_config(
        new_config=settings.settings, config_key="settings"
    )
    return ResponseMessage(message=update_config)


@app.post("/api/agent/{agent_name}/learn/file", tags=["Agent"])
async def learn_file(agent_name: str, file: FileInput) -> ResponseMessage:
    # Strip any path information from the file name
    file.file_name = os.path.basename(file.file_name)
    base_path = os.path.join(os.getcwd(), "WORKSPACE")
    file_path = os.path.normpath(os.path.join(base_path, file.file_name))
    if not file_path.startswith(base_path):
        raise Exception("Path given not allowed")
    file_content = base64.b64decode(file.file_content)
    with open(file_path, "wb") as f:
        f.write(file_content)
    try:
        memories = Agent(agent_name=agent_name).get_memories()
        await memories.mem_read_file(file_path=file_path)
        try:
            os.remove(file_path)
        except Exception:
            pass
        return ResponseMessage(message="Agent learned the content from the file.")
    except Exception as e:
        try:
            os.remove(file_path)
        except Exception:
            pass
        raise HTTPException(status_code=500, detail=str(e))


@app.post("/api/agent/{agent_name}/learn/url", tags=["Agent"])
async def learn_url(agent_name: str, url: UrlInput) -> ResponseMessage:
    try:
        memories = Agent(agent_name=agent_name).get_memories()
        await memories.read_website(url=url.url)
        return ResponseMessage(message="Agent learned the content from the url.")
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@app.put("/api/agent/{agent_name}/commands", tags=["Agent"])
async def update_agent_commands(
    agent_name: str, commands: AgentCommands
) -> ResponseMessage:
    update_config = Agent(agent_name=agent_name).update_agent_config(
        new_config=commands.commands, config_key="commands"
    )
    return ResponseMessage(message=update_config)


@app.delete("/api/agent/{agent_name}", tags=["Agent"])
async def deleteagent(agent_name: str) -> ResponseMessage:
    delete_agent(agent_name=agent_name)
    return ResponseMessage(message=f"Agent {agent_name} deleted.")


@app.get("/api/agent", tags=["Agent"])
async def getagents():
    agents = get_agents()
    return {"agents": agents}


@app.get("/api/agent/{agent_name}", tags=["Agent"])
async def get_agentconfig(agent_name: str):
    agent_config = Agent(agent_name=agent_name).get_agent_config()
    return {"agent": agent_config}


@app.get("/api/{agent_name}/chat", tags=["Agent"])
async def get_chat_history(agent_name: str):
    chat_history = Agent(agent_name=agent_name).get_history()
    return {"chat_history": chat_history}


@app.delete("/api/agent/{agent_name}/history", tags=["Agent"])
async def delete_history(agent_name: str) -> ResponseMessage:
    Agent(agent_name=agent_name).delete_history()
    return ResponseMessage(message=f"History for agent {agent_name} deleted.")


@app.delete("/api/agent/{agent_name}/history/message", tags=["Agent"])
async def delete_history_message(
    agent_name: str, message: ResponseMessage
) -> ResponseMessage:
    Agent(agent_name=agent_name).delete_history_message(message.message)
    return ResponseMessage(message=f"Message deleted.")


@app.delete("/api/agent/{agent_name}/memory", tags=["Agent"])
async def wipe_agent_memories(agent_name: str) -> ResponseMessage:
    Agent(agent_name=agent_name).wipe_agent_memories()
    return ResponseMessage(message=f"Memories for agent {agent_name} deleted.")


@app.post("/api/agent/{agent_name}/instruct", tags=["Agent"])
async def instruct(agent_name: str, prompt: Prompt):
    agent = Interactions(agent_name=agent_name)
    response = await agent.run(
        user_input=prompt.prompt,
        prompt="instruct",
    )
    return {"response": str(response)}


@app.post("/api/agent/{agent_name}/prompt", tags=["Agent"])
async def prompt_agent(agent_name: str, agent_prompt: AgentPrompt):
    agent = Interactions(agent_name=agent_name)
    response = await agent.run(
        prompt=agent_prompt.prompt_name,
        websearch=agent_prompt.websearch,
        websearch_depth=agent_prompt.websearch_depth,
        context_results=agent_prompt.context_results,
        **agent_prompt.prompt_args,
    )
    return {"response": str(response)}


@app.post("/api/agent/{agent_name}/smartinstruct/{shots}", tags=["Agent"])
async def smartinstruct(agent_name: str, shots: int, prompt: Prompt):
    agent = Interactions(agent_name=agent_name)
    response = await agent.smart_instruct(user_input=prompt.prompt, shots=int(shots))
    return {"response": str(response)}


@app.post("/api/agent/{agent_name}/chat", tags=["Agent"])
async def chat(agent_name: str, prompt: Prompt):
    agent = Interactions(agent_name=agent_name)
    response = await agent.run(
        user_input=prompt.prompt, prompt="Chat", context_results=6
    )
    return {"response": str(response)}


@app.post("/api/v1/completions", tags=["Agent"])
async def completion(prompt: Completions):
    # prompt.model is the agent name
    agent = Interactions(agent_name=prompt.model)
    agent_config = Agent(agent_name=prompt.model).get_agent_config()
    if "settings" in agent_config:
        if "AI_MODEL" in agent_config["settings"]:
            model = agent_config["settings"]["AI_MODEL"]
        else:
            model = "undefined"
    else:
        model = "undefined"
    response = await agent.run(
        user_input=prompt.prompt,
        prompt="Custom Input",
        context_results=3,
        shots=prompt.n,
    )
    characters = string.ascii_letters + string.digits
    random_chars = "".join(random.choice(characters) for _ in range(15))
    res_model = {
        "id": f"cmpl-{random_chars}",
        "object": "text_completion",
        "created": int(time.time()),
        "model": model,
        "choices": [
            {
                "text": response,
                "index": 0,
                "logprobs": None,
                "finish_reason": "stop",
            }
        ],
        "usage": {"prompt_tokens": 0, "completion_tokens": 0, "total_tokens": 0},
    }
    return res_model


@app.post("/api/v1/chat/completions", tags=["Agent"])
async def chat_completion(prompt: Completions):
    # prompt.model is the agent name
    agent = Interactions(agent_name=prompt.model)
    agent_config = Agent(agent_name=prompt.model).get_agent_config()
    if "settings" in agent_config:
        if "AI_MODEL" in agent_config["settings"]:
            model = agent_config["settings"]["AI_MODEL"]
        else:
            model = "undefined"
    else:
        model = "undefined"
    response = await agent.run(
        user_input=prompt.prompt,
        prompt="Custom Input",
        context_results=3,
        shots=prompt.n,
    )
    characters = string.ascii_letters + string.digits
    random_chars = "".join(random.choice(characters) for _ in range(15))
    res_model = {
        "id": f"chatcmpl-{random_chars}",
        "object": "chat.completion",
        "created": int(time.time()),
        "model": model,
        "choices": [
            {
                "index": 0,
                "message": [
                    {
                        "role": "assistant",
                        "content": response,
                    },
                ],
                "finish_reason": "stop",
            }
        ],
        "usage": {"prompt_tokens": 0, "completion_tokens": 0, "total_tokens": 0},
    }
    return res_model


@app.post("/api/agent/{agent_name}/smartchat/{shots}", tags=["Agent"])
async def smartchat(agent_name: str, shots: int, prompt: Prompt):
    agent = Interactions(agent_name=agent_name)
    response = await agent.smart_chat(user_input=prompt.prompt, shots=shots)
    return {"response": str(response)}


@app.get("/api/agent/{agent_name}/command", tags=["Agent"])
async def get_commands(agent_name: str):
    agent = Agent(agent_name=agent_name)
    return {"commands": agent.agent_config["commands"]}


@app.patch("/api/agent/{agent_name}/command", tags=["Agent"])
async def toggle_command(
    agent_name: str, payload: ToggleCommandPayload
) -> ResponseMessage:
    agent = Agent(agent_name=agent_name)
    try:
        if payload.command_name == "*":
            for each_command_name in agent.agent_config["commands"]:
                agent.agent_config["commands"][each_command_name] = payload.enable

            agent.update_agent_config(
                new_config=agent.agent_config["commands"], config_key="commands"
            )
            return ResponseMessage(
                message=f"All commands enabled for agent '{agent_name}'."
            )
        else:
            agent.agent_config["commands"][payload.command_name] = payload.enable
            agent.update_agent_config(
                new_config=agent.agent_config["commands"], config_key="commands"
            )
            return ResponseMessage(
                message=f"Command '{payload.command_name}' toggled for agent '{agent_name}'."
            )
    except Exception as e:
        logging.info(e)
        raise HTTPException(
            status_code=500,
            detail=f"Error enabling all commands for agent '{agent_name}': {str(e)}",
        )


@app.get("/api/chain", tags=["Chain"])
async def get_chains():
    chains = Chain().get_chains()
    return chains


@app.get("/api/chain/{chain_name}", tags=["Chain"])
async def get_chain(chain_name: str):
    try:
        chain_data = Chain().get_chain(chain_name=chain_name)
        return {"chain": chain_data}
    except:
        raise HTTPException(status_code=404, detail="Chain not found")


@app.get("/api/chain/{chain_name}/responses", tags=["Chain"])
async def get_chain_responses(chain_name: str):
    try:
        chain_data = Chain().get_step_response(chain_name=chain_name, step_number="all")
        return {"chain": chain_data}
    except:
        raise HTTPException(status_code=404, detail="Chain not found")


@app.post("/api/chain/{chain_name}/run", tags=["Chain"])
async def run_chain(chain_name: str, user_input: RunChain):
    chain_response = await Interactions(agent_name="").run_chain(
        chain_name=chain_name,
        user_input=user_input.prompt,
        agent_override=user_input.agent_override,
        all_responses=user_input.all_responses,
        from_step=user_input.from_step,
    )
    return chain_response


@app.post("/api/chain", tags=["Chain"])
async def add_chain(chain_name: ChainName) -> ResponseMessage:
    Chain().add_chain(chain_name=chain_name.chain_name)
    return ResponseMessage(message=f"Chain '{chain_name.chain_name}' created.")


@app.post("/api/chain/import", tags=["Chain"])
async def importchain(chain: ChainData) -> ResponseMessage:
    print(chain)
    response = import_chain(chain_name=chain.chain_name, steps=chain.steps)
    return ResponseMessage(message=response)


@app.put("/api/chain/{chain_name}", tags=["Chain"])
async def rename_chain(chain_name: str, new_name: ChainNewName) -> ResponseMessage:
    Chain().rename_chain(chain_name=chain_name, new_name=new_name.new_name)
    return ResponseMessage(
        message=f"Chain '{chain_name}' renamed to '{new_name.new_name}'."
    )


@app.delete("/api/chain/{chain_name}", tags=["Chain"])
async def delete_chain(chain_name: str) -> ResponseMessage:
    Chain().delete_chain(chain_name=chain_name)
    return ResponseMessage(message=f"Chain '{chain_name}' deleted.")


@app.post("/api/chain/{chain_name}/step", tags=["Chain"])
async def add_step(chain_name: str, step_info: StepInfo) -> ResponseMessage:
    Chain().add_chain_step(
        chain_name=chain_name,
        step_number=step_info.step_number,
        prompt_type=step_info.prompt_type,
        prompt=step_info.prompt,
        agent_name=step_info.agent_name,
    )
    return {"message": f"Step {step_info.step_number} added to chain '{chain_name}'."}


@app.put("/api/chain/{chain_name}/step/{step_number}", tags=["Chain"])
async def update_step(
    chain_name: str, step_number: int, chain_step: ChainStep
) -> ResponseMessage:
    Chain().update_step(
        chain_name=chain_name,
        step_number=chain_step.step_number,
        prompt_type=chain_step.prompt_type,
        prompt=chain_step.prompt,
        agent_name=chain_step.agent_name,
    )
    return {
        "message": f"Step {chain_step.step_number} updated for chain '{chain_name}'."
    }


@app.patch("/api/chain/{chain_name}/step/move", tags=["Chain"])
async def move_step(
    chain_name: str, chain_step_new_info: ChainStepNewInfo
) -> ResponseMessage:
    Chain().move_step(
        chain_name=chain_name,
        current_step_number=chain_step_new_info.old_step_number,
        new_step_number=chain_step_new_info.new_step_number,
    )
    return {
        "message": f"Step {chain_step_new_info.old_step_number} moved to {chain_step_new_info.new_step_number} in chain '{chain_name}'."
    }


@app.delete("/api/chain/{chain_name}/step/{step_number}", tags=["Chain"])
async def delete_step(chain_name: str, step_number: int) -> ResponseMessage:
    Chain().delete_step(chain_name=chain_name, step_number=step_number)
    return {"message": f"Step {step_number} deleted from chain '{chain_name}'."}


@app.post("/api/prompt", tags=["Prompt"])
async def add_prompt(prompt: CustomPromptModel) -> ResponseMessage:
    try:
        Prompts().add_prompt(prompt_name=prompt.prompt_name, prompt=prompt.prompt)
        return ResponseMessage(message=f"Prompt '{prompt.prompt_name}' added.")
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/api/prompt/{prompt_name}", tags=["Prompt"], response_model=CustomPromptModel)
async def get_prompt(prompt_name: str):
    # try:
    prompt_content = Prompts().get_prompt(prompt_name=prompt_name)
    return {"prompt_name": prompt_name, "prompt": prompt_content}
    # except Exception as e:
    #    raise HTTPException(status_code=404, detail=str(e))


@app.get("/api/prompt", response_model=PromptList, tags=["Prompt"])
async def get_prompts():
    prompts = Prompts().get_prompts()
    return {"prompts": prompts}


@app.delete("/api/prompt/{prompt_name}", tags=["Prompt"])
async def delete_prompt(prompt_name: str) -> ResponseMessage:
    try:
        Prompts().delete_prompt(prompt_name=prompt_name)
        return ResponseMessage(message=f"Prompt '{prompt_name}' deleted.")
    except Exception as e:
        raise HTTPException(status_code=404, detail=str(e))


@app.put("/api/prompt/{prompt_name}", tags=["Prompt"])
async def update_prompt(prompt: CustomPromptModel) -> ResponseMessage:
    try:
        Prompts().update_prompt(prompt_name=prompt.prompt_name, prompt=prompt.prompt)
        return ResponseMessage(message=f"Prompt '{prompt.prompt_name}' updated.")
    except Exception as e:
        raise HTTPException(status_code=404, detail=str(e))


@app.get("/api/prompt/{prompt_name}/args", tags=["Prompt"])
async def get_prompt_arg(prompt_name: str):
    return {"prompt_args": Prompts().get_prompt_args(prompt_name)}


@app.get("/api/extensions/settings", tags=["Extensions"])
async def get_extension_settings():
    try:
        return {"extension_settings": Extensions().get_extension_settings()}
    except Exception:
        raise HTTPException(status_code=400, detail="Unable to retrieve settings.")


@app.get("/api/extensions/{command_name}/args", tags=["Extension"])
async def get_command_args(command_name: str):
    return {"command_args": Extensions().get_command_args(command_name=command_name)}


@app.get("/api/extensions", tags=["Extension"])
async def get_extensions():
    return {"extensions": Extensions().get_extensions()}


if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=7437)



import streamlit as st
from components.verify_backend import verify_backend
from components.docs import agixt_docs
from streamlit_autorefresh import st_autorefresh


st.set_page_config(
    page_title="AGiXT",
    page_icon=":robot:",
    layout="wide",
    initial_sidebar_state="expanded",
)

# Run the autorefresh about (5 seconds) and stop running in the corner after the 2nd time
count = st_autorefresh(interval=5000, limit=2)


try:
    with open("./.streamlit/config.toml") as f:
        if "Dark" in f.read():
            light_theme = False
except:
    light_theme = False


agixt_docs()
if light_theme == True:
    st.markdown(
        """
        <div style="text-align: center;">
        <img src="https://josh-xt.github.io/AGiXT/images/AGiXT-gradient-flat.svg" width="65%">
        </div>
        """,
        unsafe_allow_html=True,
    )
else:
    st.markdown(
        f"""
        <div style="text-align: center;">
        <img src="https://josh-xt.github.io/AGiXT/images/AGiXT-gradient-light.svg" width="65%">
        </div>
        """,
        unsafe_allow_html=True,
    )


if not verify_backend():
    verify_backend()



import requests
import re


class OobaboogaProvider:
    def __init__(
        self,
        AI_PROVIDER_URI: str = "",
        MAX_TOKENS: int = 2048,
        DO_SAMPLE: str = "True",
        AI_TEMPERATURE: float = 0.7,
        TOP_P: float = 0.9,
        TYPICAL_P: float = 0.2,
        EPSILON_CUTOFF: float = 0,
        ETA_CUTOFF: float = 0,
        TFS: float = 1,
        TOP_A: float = 0,
        REPETITION_PENALTY: float = 1.05,
        ENCODER_REPETITION_PENALTY: float = 1,
        TOP_K: int = 200,
        MIN_LENGTH: float = 0,
        NO_REPEAT_NGRAM_SIZE: float = 0,
        NUM_BEAMS: int = 1,
        PENALTY_ALPHA: float = 0,
        LENGTH_PENALTY: float = 1,
        MIROSTAT_MODE: float = 0,
        MIROSTAT_TAU: float = 5,
        MIROSTAT_ETA: float = 0.1,
        TRUNCATION_LENGTH: int = 2048,
        AI_MODEL: str = "default",
        PROMPT_PREFIX: str = "",
        PROMPT_SUFFIX: str = "",
        **kwargs,
    ):
        self.AI_PROVIDER_URI = (
            AI_PROVIDER_URI if AI_PROVIDER_URI else "http://localhost:5000"
        )
        self.MAX_TOKENS = MAX_TOKENS if MAX_TOKENS else 2048
        self.DO_SAMPLE = DO_SAMPLE if DO_SAMPLE else "True"
        self.AI_TEMPERATURE = AI_TEMPERATURE if AI_TEMPERATURE else 0.7
        self.AI_MODEL = AI_MODEL if AI_MODEL else "default"
        self.TOP_P = TOP_P if TOP_P else 0.9
        self.TYPICAL_P = TYPICAL_P if TYPICAL_P else 0.2
        self.EPSILON_CUTOFF = EPSILON_CUTOFF if EPSILON_CUTOFF else 0
        self.ETA_CUTOFF = ETA_CUTOFF if ETA_CUTOFF else 0
        self.TFS = TFS if TFS else 1
        self.TOP_A = TOP_A if TOP_A else 0
        self.REPETITION_PENALTY = REPETITION_PENALTY if REPETITION_PENALTY else 1.05
        self.ENCODER_REPETITION_PENALTY = (
            ENCODER_REPETITION_PENALTY if ENCODER_REPETITION_PENALTY else 1
        )
        self.TOP_K = TOP_K if TOP_K else 200
        self.MIN_LENGTH = MIN_LENGTH if MIN_LENGTH else 0
        self.NO_REPEAT_NGRAM_SIZE = NO_REPEAT_NGRAM_SIZE if NO_REPEAT_NGRAM_SIZE else 0
        self.NUM_BEAMS = NUM_BEAMS if NUM_BEAMS else 1
        self.PENALTY_ALPHA = PENALTY_ALPHA if PENALTY_ALPHA else 0
        self.LENGTH_PENALTY = LENGTH_PENALTY if LENGTH_PENALTY else 1
        self.MIROSTAT_MODE = MIROSTAT_MODE if MIROSTAT_MODE else 0
        self.MIROSTAT_TAU = MIROSTAT_TAU if MIROSTAT_TAU else 5
        self.MIROSTAT_ETA = MIROSTAT_ETA if MIROSTAT_ETA else 0.1
        self.TRUNCATION_LENGTH = TRUNCATION_LENGTH if TRUNCATION_LENGTH else 2048
        self.PROMPT_PREFIX = PROMPT_PREFIX if PROMPT_PREFIX else ""
        self.PROMPT_SUFFIX = PROMPT_SUFFIX if PROMPT_SUFFIX else ""
        self.requirements = []

    async def instruct(self, prompt, tokens: int = 0):
        new_tokens = int(self.MAX_TOKENS) - tokens
        prompt = f"{self.PROMPT_PREFIX}{prompt}{self.PROMPT_SUFFIX}"
        params = {
            "prompt": prompt,
            "max_new_tokens": new_tokens,
            "do_sample": True,
            "temperature": float(self.AI_TEMPERATURE),
            "top_p": float(self.TOP_P),
            "typical_p": float(self.TYPICAL_P),
            "epsilon_cutoff": float(self.EPSILON_CUTOFF),
            "eta_cutoff": float(self.ETA_CUTOFF),
            "tfs": float(self.TFS),
            "top_a": float(self.TOP_A),
            "repetition_penalty": float(self.REPETITION_PENALTY),
            "encoder_repetition_penalty": float(self.ENCODER_REPETITION_PENALTY),
            "top_k": int(self.TOP_K),
            "min_length": float(self.MIN_LENGTH),
            "no_repeat_ngram_size": float(self.NO_REPEAT_NGRAM_SIZE),
            "num_beams": int(self.NUM_BEAMS),
            "penalty_alpha": float(self.PENALTY_ALPHA),
            "length_penalty": float(self.LENGTH_PENALTY),
            "early_stopping": False,
            "mirostat_mode": float(self.MIROSTAT_MODE),
            "mirostat_tau": float(self.MIROSTAT_TAU),
            "mirostat_eta": float(self.MIROSTAT_ETA),
            "seed": -1,
            "add_bos_token": True,
            "truncation_length": int(self.TRUNCATION_LENGTH),
            "ban_eos_token": False,
            "skip_special_tokens": True,
            "custom_stopping_strings": "",  # leave this blank
            "stopping_strings": ["</s>"],
        }
        print(params)
        response = requests.post(f"{self.AI_PROVIDER_URI}/api/v1/generate", json=params)
        data = None

        if response.status_code == 200:
            data = response.json()["results"][0]["text"]
            data = re.sub(r"(?<!\\)\\(?!n)", "", data)
        return data



from ApiClient import ApiClient
import streamlit as st
import html
import re


def get_history(agent_name):
    st.markdown("### Agent History")
    st.markdown(
        "The history of the agent's interactions. The latest responses are at the top."
    )

    # Add a button to delete agent history
    if st.button("Delete Agent History"):
        ApiClient.delete_agent_history(agent_name=agent_name)
        st.success("Agent history deleted successfully.")

    # Define CSS rules for message container
    message_container_css = """
        <style>
        .message-container {
            height: 400px;
            overflow: auto;
            border: 1px solid #ddd;
            border-radius: 5px;
            padding: 10px;
            margin: 10px 0;
        }
        .message {
            margin-bottom: 10px;
        }
        .user-message {
            background-color: #0f0f0f;
            padding: 5px;
            border-radius: 5px;
        }
        .agent-message {
            background-color: #3a3b3c;
            color: white;
            padding: 5px;
            border-radius: 5px;
        }
        </style>
    """

    st.write(message_container_css, unsafe_allow_html=True)

    with st.container():
        message_container = "<div class='message-container'>"
        try:
            history = ApiClient.get_chat_history(agent_name=agent_name)
            history = reversed(history)

            for item in history:
                item["message"] = html.escape(item["message"])
                item["message"] = item["message"].replace(r"\n", "<br>")

                # Check if the message is a code block
                code_block_match = re.search(r"```(.*)```", item["message"], re.DOTALL)

                if code_block_match:
                    # Extract the code block from the match object
                    code_message = code_block_match.group(1)
                    # Replace the code block in the original message with the code block wrapped in <pre><code> tags
                    item["message"] = re.sub(
                        r"```.*```",
                        f"<pre><code>{code_message}</code></pre>",
                        item["message"],
                        flags=re.DOTALL,
                    )

                message = f"{item['timestamp']}<br><b>{item['role']}:</b><br>{item['message']}"

                if agent_name in item["role"]:
                    message_container += (
                        f"<div class='message agent-message'>{message}</div>"
                    )
                else:
                    message_container += (
                        f"<div class='message user-message'>{message}</div>"
                    )

        except Exception as e:
            print(e)
        message_container += "</div>"
        st.write(message_container, unsafe_allow_html=True)



import os
import json


def create_command_suggestion_chain(agent_name, command_name, command_args):
    chain = Chain()
    chains = chain.get_chains()
    chain_name = f"{agent_name} Command Suggestions"
    if chain_name in chains:
        step = int(chain.get_chain(chain_name=chain_name)["steps"][-1]["step"]) + 1
    else:
        chain.add_chain(chain_name=chain_name)
        step = 1
    chain.add_chain_step(
        chain_name=chain_name,
        agent_name=agent_name,
        step_number=step,
        prompt_type="Command",
        prompt={
            "command_name": command_name,
            **command_args,
        },
    )
    return f"The command has been added to a chain called '{agent_name} Command Suggestions' for you to review and execute manually."


def get_chain_file_path(chain_name):
    base_path = os.path.join(os.getcwd(), "chains")
    folder_path = os.path.normpath(os.path.join(base_path, chain_name))
    file_path = os.path.normpath(os.path.join(base_path, f"{chain_name}.json"))
    if not file_path.startswith(base_path) or not folder_path.startswith(base_path):
        raise ValueError("Invalid path, chain name must not contain slashes.")
    if not os.path.exists(folder_path):
        os.mkdir(folder_path)
    return file_path


def get_chain_responses_file_path(chain_name):
    base_path = os.path.join(os.getcwd(), "chains")
    file_path = os.path.normpath(os.path.join(base_path, chain_name, "responses.json"))
    if not file_path.startswith(base_path):
        raise ValueError("Invalid path, chain name must not contain slashes.")
    return file_path


def import_chain(chain_name: str, steps: dict):
    file_path = get_chain_file_path(chain_name=chain_name)
    steps = steps["steps"] if "steps" in steps else steps
    with open(file_path, "w") as f:
        json.dump({"chain_name": chain_name, "steps": steps}, f)
    return f"Chain '{chain_name}' imported."


class Chain:
    def get_chain(self, chain_name):
        try:
            file_path = get_chain_file_path(chain_name=chain_name)
            with open(file_path, "r") as f:
                chain_data = json.load(f)
            return chain_data
        except:
            return {}

    def get_chains(self):
        chains = [
            f.replace(".json", "") for f in os.listdir("chains") if f.endswith(".json")
        ]
        return chains

    def add_chain(self, chain_name):
        file_path = get_chain_file_path(chain_name=chain_name)
        chain_data = {"chain_name": chain_name, "steps": []}
        with open(file_path, "w") as f:
            json.dump(chain_data, f)

    def rename_chain(self, chain_name, new_name):
        file_path = get_chain_file_path(chain_name=chain_name)
        new_file_path = get_chain_file_path(chain_name=new_name)
        os.rename(
            os.path.join(file_path),
            os.path.join(new_file_path),
        )
        chain_data = self.get_chain(chain_name=new_name)
        chain_data["chain_name"] = new_name
        with open(new_file_path, "w") as f:
            json.dump(chain_data, f)

    def add_chain_step(self, chain_name, step_number, agent_name, prompt_type, prompt):
        file_path = get_chain_file_path(chain_name=chain_name)
        chain_data = self.get_chain(chain_name=chain_name)
        chain_data["steps"].append(
            {
                "step": step_number,
                "agent_name": agent_name,
                "prompt_type": prompt_type,
                "prompt": prompt,
            }
        )
        with open(file_path, "w") as f:
            json.dump(chain_data, f)

    def update_step(self, chain_name, step_number, agent_name, prompt_type, prompt):
        file_path = get_chain_file_path(chain_name=chain_name)
        chain_data = self.get_chain(chain_name=chain_name)
        for step in chain_data["steps"]:
            if step["step"] == step_number:
                step["agent_name"] = agent_name
                step["prompt_type"] = prompt_type
                step["prompt"] = prompt
                break
        with open(file_path, "w") as f:
            json.dump(chain_data, f)

    def delete_step(self, chain_name, step_number):
        file_path = get_chain_file_path(chain_name=chain_name)
        chain_data = self.get_chain(chain_name=chain_name)
        chain_data["steps"] = [
            step for step in chain_data["steps"] if step["step"] != step_number
        ]
        with open(file_path, "w") as f:
            json.dump(chain_data, f)

    def delete_chain(self, chain_name):
        file_path = get_chain_file_path(chain_name=chain_name)
        os.remove(file_path)

    def get_step(self, chain_name, step_number):
        chain_data = self.get_chain(chain_name=chain_name)
        for step in chain_data["steps"]:
            if step["step"] == step_number:
                return step
        return None

    def get_steps(self, chain_name):
        chain_data = self.get_chain(chain_name=chain_name)
        return chain_data["steps"]

    def move_step(self, chain_name, current_step_number, new_step_number):
        file_path = get_chain_file_path(chain_name=chain_name)
        chain_data = self.get_chain(chain_name=chain_name)
        if not 1 <= new_step_number <= len(
            chain_data["steps"]
        ) or current_step_number not in [step["step"] for step in chain_data["steps"]]:
            print(f"Error: Invalid step numbers.")
            return
        moved_step = None
        for step in chain_data["steps"]:
            if step["step"] == current_step_number:
                moved_step = step
                chain_data["steps"].remove(step)
                break
        for step in chain_data["steps"]:
            if new_step_number < current_step_number:
                if new_step_number <= step["step"] < current_step_number:
                    step["step"] += 1
            else:
                if current_step_number < step["step"] <= new_step_number:
                    step["step"] -= 1
        moved_step["step"] = new_step_number
        chain_data["steps"].append(moved_step)
        chain_data["steps"] = sorted(chain_data["steps"], key=lambda x: x["step"])
        with open(file_path, "w") as f:
            json.dump(chain_data, f)

    def get_step_response(self, chain_name, step_number="all"):
        file_path = get_chain_responses_file_path(chain_name=chain_name)
        try:
            with open(file_path, "r") as f:
                responses = json.load(f)
            if step_number == "all":
                return responses
            else:
                return responses.get(str(step_number))
        except:
            return ""

    def get_chain_responses(self, chain_name):
        file_path = get_chain_responses_file_path(chain_name=chain_name)
        try:
            with open(file_path, "r") as f:
                responses = json.load(f)
            return responses
        except:
            return {}



from ApiClient import ApiClient
import streamlit as st
import os
import logging


@st.cache_data
def cached_get_extensions():
    return ApiClient.get_extensions()


@st.cache_data
def cached_get_prompts():
    return ApiClient.get_prompts()


def build_args(args: dict = {}, prompt: dict = {}, step_number: int = 0):
    return {
        arg: st.text_input(arg, value=prompt.get(arg, ""), key=f"{arg}_{step_number}")
        for arg in args
        if arg != "context"
        and arg != "command_list"
        and arg != "COMMANDS"
        and arg != "user_input"
    }


def prompt_selection(prompt: dict = {}, step_number: int = 0):
    available_prompts = cached_get_prompts()
    prompt_name = st.selectbox(
        "Select Custom Prompt",
        [""] + available_prompts,
        index=available_prompts.index(prompt.get("prompt_name", "")) + 1
        if "prompt_name" in prompt
        else 0,
        key=f"step_{step_number}_prompt_name",
    )

    if prompt_name:
        prompt_args = ApiClient.get_prompt_args(prompt_name)
        args = build_args(args=prompt_args, prompt=prompt, step_number=step_number)
        new_prompt = {
            "prompt_name": prompt_name,
            **args,
        }
        return new_prompt


def command_selection(prompt: dict = {}, step_number: int = 0):
    agent_commands = cached_get_extensions()
    available_commands = [cmd[0] for cmd in agent_commands]
    command_name = st.selectbox(
        "Select Command",
        [""] + available_commands,
        key=f"command_name_{step_number}",
        index=available_commands.index(prompt.get("command_name", "")) + 1
        if "command_name" in prompt
        else 0,
    )

    if command_name:
        command_args = ApiClient.get_command_args(command_name=command_name)
        args = build_args(args=command_args, prompt=prompt, step_number=step_number)
        new_prompt = {
            "command_name": command_name,
            **args,
        }
        return new_prompt


def chain_selection(prompt: dict = {}, step_number: int = 0):
    available_chains = ApiClient.get_chains()
    chain_name = st.selectbox(
        "Select Chain",
        [""] + available_chains,
        index=available_chains.index(prompt.get("chain_name", "")) + 1
        if "chain_name" in prompt
        else 0,
        key=f"step_{step_number}_chain_name",
    )
    user_input = st.text_input(
        "User Input",
        value=prompt.get("user_input", ""),
        key=f"user_input_{step_number}",
    )

    if chain_name:
        new_prompt = {"chain_name": chain_name, "user_input": user_input}
        return new_prompt


def agent_selection():
    # Load the previously selected agent name
    try:
        with open(os.path.join("session.txt"), "r") as f:
            previously_selected_agent = f.read().strip()
    except FileNotFoundError:
        previously_selected_agent = None

    # Get the list of agent names
    agent_names = [agent["name"] for agent in ApiClient.get_agents()]

    # If the previously selected agent is in the list, use it as the default
    if previously_selected_agent in agent_names:
        default_index = (
            agent_names.index(previously_selected_agent) + 1
        )  # add 1 for the empty string at index 0
    else:
        default_index = 0

    # Create the selectbox
    selected_agent = st.selectbox(
        "Agent Name",
        options=[""] + agent_names,
        index=default_index,
        key="select_learning_agent",
    )

    # If the selected agent has changed, save the new selection
    if selected_agent != previously_selected_agent:
        with open(os.path.join("session.txt"), "w") as f:
            f.write(selected_agent)
        try:
            st.experimental_rerun()
        except Exception as e:
            logging.info(e)
    return selected_agent
